import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants


class TestDbusModuleCalls(unittest.TestCase):
    """
    Testing things that use the Dbus module
    """
    def setUp(self):
        """
        Patch the DBus module
        :return:
        """
        self.dbus_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
        }
        self.dbus_mock.Interface.return_value.GetManagedObjects.return_value = tests.obj_data.full_ubits
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import tools
        self.module_under_test = tools

    def tearDown(self):
        self.module_patcher.stop()

    def test_NoSuffix(self):
        self.assertEqual(
            self.module_under_test.url_to_advert(
                'http://camjam.me/', 0x10, 0x00),
            [0x10, 0x00, 0x02, 0x63, 0x61, 0x6D, 0x6A,
             0x61, 0x6D, 0x2E, 0x6D, 0x65, 0x2F])

    def test_WithSuffix(self):
        self.assertEqual(
            self.module_under_test.url_to_advert(
                'https://www.google.com', 0x10, 0x00),
            [0x10, 0x00, 0x01, 0x67, 0x6f,
             0x6f, 0x67, 0x6c, 0x65, 0x07])

    def test_PostSuffix(self):
        self.assertEqual(
            self.module_under_test.url_to_advert(
                'http://www.csr.com/about', 0x10, 0x00),
            [0x10, 0x00, 0x00, 0x63, 0x73, 0x72,
             0x00, 0x61, 0x62, 0x6f, 0x75, 0x74])

    def test_IntToUint32_with_zeros(self):
        little_endian = self.module_under_test.int_to_uint32(2094)
        self.assertListEqual(little_endian, [0x2E, 0x08, 0x00, 0x00])

    def test_IntToUint32_all_full(self):
        little_endian = self.module_under_test.int_to_uint32(305419896)
        self.assertListEqual(little_endian, [0x78, 0x56, 0x34, 0x12])

    def test_not_signed(self):
        self.assertListEqual([0xBB, 0xBB, 0xBB, 0xBB],
                             self.module_under_test.int_to_uint32(3149642683))
    def test_p0(self):
        self.assertListEqual([1, 0x0, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**0))

    def test_p1(self):
        self.assertListEqual([2, 0x0, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**1))

    def test_p2(self):
        self.assertListEqual([4, 0x00, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**2))

    def test_p3(self):
        self.assertListEqual([8, 0x00, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**3))

    def test_p4(self):
        self.assertListEqual([0x10, 0x00, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**4))

    def test_p5(self):
        self.assertListEqual([0x20, 0x00, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**5))

    def test_p6(self):
        self.assertListEqual([0x40, 0x00, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**6))

    def test_p7(self):
        self.assertListEqual([0x80, 0x00, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**7))

    def test_p8(self):
        self.assertListEqual([0x00, 0x01, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**8))

    def test_p9(self):
        self.assertListEqual([0, 0x02, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**9))

    def test_p10(self):
        self.assertListEqual([0, 0x04, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**10))

    def test_p11(self):
        self.assertListEqual([0, 0x08, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**11))

    def test_p12(self):
        self.assertListEqual([0, 0x10, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**12))

    def test_p13(self):
        self.assertListEqual([0, 0x20, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**13))

    def test_p14(self):
        self.assertListEqual([0, 0x40, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**14))

    def test_p15(self):
        self.assertListEqual([0, 0x80, 0x00, 0x00],
                             self.module_under_test.int_to_uint32(2**15))

    def test_p16(self):
        self.assertListEqual([0, 0x00, 0x01, 0x00],
                             self.module_under_test.int_to_uint32(2**16))

    def test_p17(self):
        self.assertListEqual([0, 0x00, 0x02, 0x00],
                             self.module_under_test.int_to_uint32(2**17))

    def test_p18(self):
        self.assertListEqual([0, 0x00, 0x04, 0x00],
                             self.module_under_test.int_to_uint32(2**18))

    def test_p19(self):
        self.assertListEqual([0, 0x00, 0x08, 0x00],
                             self.module_under_test.int_to_uint32(2**19))

    def test_IntToUint16(self):
        little_endian = self.module_under_test.int_to_uint16(43733)
        self.assertListEqual(little_endian, [0xD5, 0xAA])

    def test_Sint16ToInt(self):
        result = self.module_under_test.int_to_uint16(0b1111111011101111)
        self.assertEqual(result, [0xEF, 0xFE])

    def test_bytes_to_xyz(self):
        # result = self.module_under_test.bytes_to_xyz([32, 176, 40,239, 96, 84])
        result = self.module_under_test.bytes_to_xyz([0x20, 0x00, 0xD0, 0x00, 0x20, 0xFC])
        self.assertEqual(result, [0.032, 0.208, -0.992])


if __name__ == '__main__':
    unittest.main()



import dbus
import dbusmock
import io
from pathlib import Path
from pprint import pprint
import subprocess
from unittest import mock
from gi.repository import GLib

from tests.mock_eventloop import MockAsync, run_pending_events

# Module under test
from examples import eddystone_url_beacon


class TestAdapterExample(dbusmock.DBusTestCase):
    """
    Test mocking bluetoothd
    """

    @classmethod
    def setUpClass(cls):
        here = Path(__file__).parent
        template = str(here.joinpath('dbusmock_templates', 'bluez_scan.py'))
        cls.start_system_bus()
        cls.dbus_con = cls.get_dbus(True)
        (cls.p_mock, cls.obj_bluez) = cls.spawn_server_template(
            template, {}, stdout=subprocess.PIPE)

    def setUp(self):
        self.obj_bluez.Reset()
        self.dbusmock = dbus.Interface(self.obj_bluez, dbusmock.MOCK_IFACE)
        self.dbusmock_bluez = dbus.Interface(self.obj_bluez, 'org.bluez.Mock')

    @classmethod
    def tearDownClass(cls):
        cls.stop_dbus(cls.system_bus_pid)
        cls.p_mock.terminate()
        cls.p_mock.wait()

    def test_eddystone_url_beacon(self):
        # TODO: A very light test at the moment. Needs to do more checking
        path = self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        # self.mngr = dbus.Interface(self.dbus_con.get_object('org.bluez', '/'),
        #                            'org.freedesktop.DBus.ObjectManager')
        # pprint(self.mngr.GetManagedObjects())

        dongle = self.dbus_con.get_object('org.bluez', path)
        self.assertFalse(dongle.Get('org.bluez.Adapter1', 'Discoverable'))
        with mock.patch('bluezero.async_tools.EventLoop', MockAsync):
            eddystone_url_beacon.main()
        # print(self.dbusmock_bluez.GetCalls())

        self.assertTrue(dongle.Get('org.bluez.Adapter1', 'Discoverable'))
        # pprint(self.mngr.GetManagedObjects())



"""
This is a simple API for reading data from a micro:bit.

You will need the Bluetooth services on the micro:bit exposed.

This code was developed using the 'Bluetooth Most Services, No Security'
micro:bit hex file from:
http://www.bittysoftware.com/downloads.html
The hex file called "For micro:bit Blue - Main Bluetooth services,
pairing not required" was used.

The following link is a good reference for Bluetooth on the microbit
http://bluetooth-mdw.blogspot.co.uk/p/bbc-microbit.html
"""
from time import sleep

from bluezero import central
from bluezero import tools
from bluezero import constants

ACCEL_SRV = 'E95D0753-251D-470A-A062-FA1922DFA9A8'
ACCEL_DATA = 'E95DCA4B-251D-470A-A062-FA1922DFA9A8'
ACCEL_PERIOD = 'E95DFB24-251D-470A-A062-FA1922DFA9A8'
BTN_SRV = 'E95D9882-251D-470A-A062-FA1922DFA9A8'
BTN_A_STATE = 'E95DDA90-251D-470A-A062-FA1922DFA9A8'
BTN_B_STATE = 'E95DDA91-251D-470A-A062-FA1922DFA9A8'
LED_SRV = 'E95DD91D-251D-470A-A062-FA1922DFA9A8'
LED_STATE = 'E95D7B77-251D-470A-A062-FA1922DFA9A8'
LED_TEXT = 'E95D93EE-251D-470A-A062-FA1922DFA9A8'
LED_SCROLL = 'E95D0D2D-251D-470A-A062-FA1922DFA9A8'
MAGNETO_SRV = 'E95DF2D8-251D-470A-A062-FA1922DFA9A8'
MAGNETO_DATA = 'E95DFB11-251D-470A-A062-FA1922DFA9A8'
MAGNETO_PERIOD = 'E95D386C-251D-470A-A062-FA1922DFA9A8'
MAGNETO_BEARING = 'E95D9715-251D-470A-A062-FA1922DFA9A8'
MAGNETO_CALIBRATE = 'E95DB358-251D-470A-A062-FA1922DFA9A8'
IO_PIN_SRV = 'E95D127B-251D-470A-A062-FA1922DFA9A8'
IO_PIN_DATA = 'E95D8D00-251D-470A-A062-FA1922DFA9A8'
IO_AD_CONFIG = 'E95D5899-251D-470A-A062-FA1922DFA9A8'
IO_PIN_CONFIG = 'E95DB9FE-251D-470A-A062-FA1922DFA9A8'
IO_PIN_PWM = 'E95DD822-251D-470A-A062-FA1922DFA9A8'
TEMP_SRV = 'E95D6100-251D-470A-A062-FA1922DFA9A8'
TEMP_DATA = 'E95D9250-251D-470A-A062-FA1922DFA9A8'
TEMP_PERIOD = 'E95D1B25-251D-470A-A062-FA1922DFA9A8'
UART_SRV = '6E400001-B5A3-F393-E0A9-E50E24DCCA9E'
UART_TX = '6E400002-B5A3-F393-E0A9-E50E24DCCA9E'
UART_RX = '6E400003-B5A3-F393-E0A9-E50E24DCCA9E'

MICROBIT = {ACCEL_SRV: [ACCEL_DATA, ACCEL_PERIOD],
            BTN_SRV: [BTN_A_STATE, BTN_B_STATE],
            LED_SRV: [LED_STATE, LED_TEXT, LED_SCROLL],
            MAGNETO_SRV: [MAGNETO_DATA, MAGNETO_PERIOD,
                          MAGNETO_BEARING, MAGNETO_CALIBRATE],
            IO_PIN_SRV: [IO_PIN_DATA, IO_AD_CONFIG, IO_PIN_CONFIG, IO_PIN_PWM],
            TEMP_SRV: [TEMP_DATA, TEMP_PERIOD],
            UART_SRV: [UART_TX, UART_RX]}
SERVICE_NAMES = {ACCEL_SRV: 'Accelerometer Service',
                 BTN_SRV: 'Button Service',
                 LED_SRV: 'LED Service',
                 MAGNETO_SRV: 'Magnetometer Service',
                 IO_PIN_SRV: 'IO Pin Service',
                 TEMP_SRV: 'Temperature Service',
                 UART_SRV: 'Nordic Semiconductor UART service',
                 }
logger = tools.create_module_logger(__name__)


class Microbit:
    """
    Class to simplify interacting with a micro:bit over Bluetooth Low Energy
    """
    def __init__(self, device_addr, adapter_addr=None, **kwargs):
        """
        Initialization of an instance of a remote micro:bit

        :param device_addr: Discovered microbit device with this address
        :param adapter_addr: Optional unless you have more than one adapter
                             on your machine
        """
        legacy_params = ['accelerometer_service', 'button_service',
                         'led_service', 'magnetometer_service',
                         'pin_service', 'temperature_service',
                         'uart_service']
        for kwarg in kwargs:
            if kwarg in legacy_params:
                logger.warning('The parameter %s has been deprecated. There '
                               'is no longer a requirement to specify which '
                               'services the micro:bit has.\nYou will get an '
                               'Error in the log if you access a '
                               'characteristic that does not exist on '
                               'micro:bit')
        self.ubit = central.Central(adapter_addr=adapter_addr,
                                    device_addr=device_addr)

        self.user_pin_callback = None
        self.user_btn_a_callback = None
        self.user_btn_b_callback = None
        self.user_calibrate_cb = None
        self.uart_tx_cb = None
        self.user_accel_cb = None

        # Micro:bit Characteristics
        # if accelerometer_service:
        self._accel_data = self.ubit.add_characteristic(ACCEL_SRV,
                                                        ACCEL_DATA)
        self._accel_period = self.ubit.add_characteristic(ACCEL_SRV,
                                                          ACCEL_PERIOD)
        # if button_service:
        self._btn_a_state = self.ubit.add_characteristic(BTN_SRV,
                                                         BTN_A_STATE)
        self._btn_b_state = self.ubit.add_characteristic(BTN_SRV,
                                                         BTN_B_STATE)
        # if led_service:
        self._led_state = self.ubit.add_characteristic(LED_SRV,
                                                       LED_STATE)
        self._led_text = self.ubit.add_characteristic(LED_SRV,
                                                      LED_TEXT)
        self._led_scroll = self.ubit.add_characteristic(LED_SRV,
                                                        LED_SCROLL)
        # if magnetometer_service:
        self._magneto_data = self.ubit.add_characteristic(
            MAGNETO_SRV,
            MAGNETO_DATA)
        self._magneto_period = self.ubit.add_characteristic(
            MAGNETO_SRV,
            MAGNETO_PERIOD)
        self._magneto_bearing = self.ubit.add_characteristic(
            MAGNETO_SRV,
            MAGNETO_BEARING)
        self._magneto_calibrate = self.ubit.add_characteristic(
            MAGNETO_SRV,
            MAGNETO_CALIBRATE)
        # if pin_service:
        self._io_pin_data = self.ubit.add_characteristic(IO_PIN_SRV,
                                                         IO_PIN_DATA)
        self._io_ad_config = self.ubit.add_characteristic(IO_PIN_SRV,
                                                          IO_AD_CONFIG)
        self._io_pin_config = self.ubit.add_characteristic(IO_PIN_SRV,
                                                           IO_PIN_CONFIG)
        self._io_pin_pwm = self.ubit.add_characteristic(IO_PIN_SRV,
                                                        IO_PIN_PWM)
        # if temperature_service:
        self._temp_data = self.ubit.add_characteristic(TEMP_SRV,
                                                       TEMP_DATA)
        self._temp_period = self.ubit.add_characteristic(TEMP_SRV,
                                                         TEMP_PERIOD)
        # if uart_service:
        self._uart_tx = self.ubit.add_characteristic(UART_SRV,
                                                     UART_TX)
        self._uart_rx = self.ubit.add_characteristic(UART_SRV,
                                                     UART_RX)

    @staticmethod
    def available(adapter_address=None):
        """
        Generator to find already discovered micro:bit's and return a tuple
        of their address and name

        :param adapter_address: Optional if you want o find devices on specific
         adapter
        :return: Microbit Object
        """
        for ubit in central.Central.available(adapter_address):
            if ubit.name and 'micro:bit' in ubit.name:
                yield Microbit(device_addr=ubit.address,
                               adapter_addr=ubit.adapter)

    def services_available(self):
        """Return a list of service names available on this micro:bit"""
        named_services = []
        for service in self.ubit.services_available:
            if service.upper() in SERVICE_NAMES:
                named_services.append(SERVICE_NAMES[service.upper()])
        return named_services

    @property
    def name(self):
        """Read the devices Bluetooth address"""
        return self.ubit.rmt_device.name

    @property
    def address(self):
        """Read the devices Bluetooth address"""
        return self.ubit.rmt_device.address

    @property
    def connected(self):
        """Indicate whether the remote device is currently connected."""
        return self.ubit.connected

    def connect(self):
        """
        Connect to the specified micro:bit for this instance
        """
        self.ubit.connect()

    def disconnect(self):
        """
        Disconnect from the micro:bit
        """
        self.ubit.disconnect()

    @property
    def scroll_delay(self):
        """
        Specifies a millisecond delay to wait for in between showing each
        character on the display.
        """
        return int.from_bytes(self._led_scroll.value,
                              byteorder='little',
                              signed=False)

    @scroll_delay.setter
    def scroll_delay(self, delay=None):
        """
        Specifies a millisecond delay to wait for in between showing each
        character on the display.
        """
        if delay < 0:
            delay = 0
        elif delay > 2**16:
            delay = 2**16
        self._led_scroll.value = tools.int_to_uint16(delay)

    @property
    def text(self):
        """
        Specify text to be displayed. Limit of 20 characters.
        The content will be restricted to that number of characters.

        :param words:
        """
        pass

    @text.setter
    def text(self, words):
        """
        Specify text to be displayed. Limit of 20 characters.
        The content will be restricted to that number of characters.

        :param words:
        """
        data = []
        text = ''
        if len(words) > 20:
            text = words[:19]
        else:
            text = words
        for letter in text:
            data.append(ord(letter))
        self._led_text.value = data

    def _write_pixels(self, data):
        """
        Utility function for the different display functions

        :param data: list of 5 numbers in the range 0 to 255
        (e.g. [0xff, 0x00, 0, 255, 0b10101]
        """

        self._led_state.value = data

    def clear_display(self):
        """
        Clear the LED display on the microbit
        """
        self._write_pixels([0x00, 0x00, 0x00, 0x00, 0x00])

    @property
    def pixels(self):
        """
        Returns a list of 5 binary numbers. Each number represents a row
        from top to bottom

        :return: Example [0b01110, 0b01000, 0b10000, 0b10000, 0b01110]
        """
        rows = self._led_state.value
        return [int(i) for i in rows]

    @pixels.setter
    def pixels(self, rows):
        """
        For each row of LEDs specify which LEDs will be on.
        :example:
        0b11111 will turn all LEDs in specified row on
        0b10101 will turn alternate LEDs on
        0b00000 will turn all LEDs in row off

        :param row0: top row
        :param row1:
        :param row2: middle row
        :param row3:
        :param row4: bottom row
        """
        self._write_pixels([rows[0], rows[1], rows[2], rows[3], rows[4]])

    @property
    def temperature(self):
        """
        Temperature from sensors in micro:bit processors

        :return: Integer of temperature in Celsius
        """
        tmp_val = self._temp_data.value

        return int.from_bytes(tmp_val, byteorder='little', signed=True)

    @property
    def button_a(self):
        """
        Read the state of button A on a micro:bit
        3 button states are defined and represented by a simple numeric
        enumeration:  0 = not pressed, 1 = pressed, 2 = long press.

        :return: integer representing button value
        """
        btn_val = self._btn_a_state.value

        return int.from_bytes(btn_val, byteorder='little', signed=False)

    @property
    def button_b(self):
        """
        Read the state of button B on a micro:bit
        3 button states are defined and represented by a simple numeric
        enumeration:  0 = not pressed, 1 = pressed, 2 = long press.

        :return: integer representing button value
        """
        btn_val = self._btn_b_state.value

        return int.from_bytes(btn_val, byteorder='little', signed=False)

    def _decode_btn_a(self, *button_values):
        """Decode button A state and pass on to user callback."""
        if 'Value' in button_values[1]:
            self.user_btn_a_callback(int(button_values[1]['Value'][0]))

    def _decode_btn_b(self, *button_values):
        """Decode button B state and pass on to user callback."""
        if 'Value' in button_values[1]:
            self.user_btn_b_callback(int(button_values[1]['Value'][0]))

    def subscribe_button_a(self, user_callback):
        """
        Execute user_callback on Button A being press on micro:bit

        :param user_callback: User callback method receiving the button state
        :return:
        """
        self.user_btn_a_callback = user_callback
        self._btn_a_state.add_characteristic_cb(self._decode_btn_a)
        self._btn_a_state.start_notify()

    def subscribe_button_b(self, user_callback):
        """
        Execute user_callback on Button B being press on micro:bit

        :param user_callback: User callback method receiving the button state
        :return:
        """
        self.user_btn_b_callback = user_callback
        self._btn_b_state.add_characteristic_cb(self._decode_btn_b)
        self._btn_b_state.start_notify()

    def _decode_pins(self, *pin_values):
        if pin_values[0] != 'org.bluez.GattCharacteristic1':
            return
        if 'Value' in pin_values[1]:
            self.user_pin_callback(int(pin_values[1]['Value'][0]),
                                   int(pin_values[1]['Value'][1]))

    def subscribe_pins(self, user_callback):
        """
        Execute user_callback on input pin being changed on micro:bit

        :param user_callback:
        :return:
        """
        self.user_pin_callback = user_callback
        self._io_pin_data.add_characteristic_cb(self._decode_pins)
        self._io_pin_data.start_notify()

    @property
    def accelerometer(self):
        """
        Read the values of the accelerometer on the microbit

        :return: return a list in the order of x, y & z
        """
        # [16, 0, 64, 0, 32, 252]
        # x=0.16, y=0.024, z=-0.992
        accel_bytes = self._accel_data.value

        return tools.bytes_to_xyz(accel_bytes)

    def _decode_accel(self, *accel_values):
        """Decode accelerometer values and pass on to user callback."""
        self.user_accel_cb(*tools.bytes_to_xyz(accel_values[1]['Value']))

    def subscribe_accelerometer(self, user_callback):
        """
        Execute user_callback on data being received on accelerometer service

        :param user_callback: User callback method receiving the accelerometer
            values
        :return:
        """
        self.user_accel_cb = user_callback
        self._accel_data.add_characteristic_cb(self._decode_accel)
        self._accel_data.start_notify()

    @property
    def magnetometer(self):
        """
        Exposes magnetometer data.
        A magnetometer measures a magnetic field such
        as the earth's magnetic field in 3 axes.

        :return: List of x, y & z value
        """
        mag_bytes = self._magneto_data.value

        return tools.bytes_to_xyz(mag_bytes)

    @property
    def bearing(self):
        """
        Compass bearing in degrees from North.

        :return: degrees in integer
        """
        mag_bear_val = self._magneto_bearing.value

        return int.from_bytes(mag_bear_val,
                              byteorder='little', signed=False)

    def calibrate(self):
        """
        Request a calibration of the magnetometer
        """
        self._magneto_calibrate.value = 0x01

    def subscribe_calibrate(self, user_callback):
        """
        Execute user_callback when calibration of magnetometer is complete

        :param user_callback:
        :return:
        """
        self.user_calibrate_cb = user_callback
        self._magneto_calibrate.add_characteristic_cb(self._magneto_cal_cb)
        self._magneto_calibrate.start_notify()

    def _magneto_cal_cb(self, iface, changed_props, invalidated_props):
        if iface != 'org.bluez.GattCharacteristic1':
            return
        if 'Value' in changed_props:
            self.user_calibrate_cb(int(changed_props['Value'][0]))

    def set_pin(self, pin_number, pin_input, pin_analogue):
        """
        For a given pin, set the direction and type for the microbit pin.

        :param pin_number: Pin number of the microbit
        :param pin_input: False for output, True for input
        :param pin_analogue: False for digital, True for analogue
        :return:
        """
        pin_bit = tools.int_to_uint32(2 ** pin_number)
        current_io_setting = self._pin_config
        current_ad_setting = self._pin_ad_config
        if pin_input:
            new_setting = tools.bitwise_or_2lists(pin_bit, current_io_setting)
            self._pin_config = new_setting
        else:
            pin_mask = tools.bitwise_xor_2lists(pin_bit,
                                                [0xff, 0xff, 0xff, 0xff])
            new_setting = tools.bitwise_and_2lists(pin_mask,
                                                   current_io_setting)
            self._pin_config = new_setting
        if pin_analogue:
            new_setting = tools.bitwise_or_2lists(pin_bit,
                                                  current_ad_setting)
            self._pin_ad_config = new_setting
        else:
            pin_mask = tools.bitwise_xor_2lists(pin_bit,
                                                [0xff, 0xff, 0xff, 0xff])
            new_setting = tools.bitwise_and_2lists(pin_mask,
                                                   current_ad_setting)
            self._pin_ad_config = new_setting

    @property
    def _pin_config(self):
        """
        A bit mask (32 bit) which defines which inputs will be read.
        A value of 0 means configured for output and 1 means configured
        for input.
        """
        return_val = []
        current_setting = self._io_pin_config.value
        for i in range(len(current_setting)):
            return_val.append(int(current_setting[i]))
        return return_val

    @_pin_config.setter
    def _pin_config(self, states):
        """
        A bit mask (32 bit) which defines which inputs will be read.
        A value of 0 means configured for output and 1 means configured
        for input.
        """
        self._io_pin_config.value = states

    @property
    def _pin_ad_config(self):
        """
        A bit mask (32 bit) which allows each pin to be configured for
        analogue or digital use.
        A value of 0 means digital and 1 means analogue.
        If no states are specified then the current state is returned
        """
        return_val = []
        current_setting = self._io_ad_config.value
        for i in range(len(current_setting)):
            return_val.append(int(current_setting[i]))
        return return_val

    @_pin_ad_config.setter
    def _pin_ad_config(self, states):
        """
        A bit mask (32 bit) which allows each pin to be configured for
        analogue or digital use.
        A value of 0 means digital and 1 means analogue.
        If no states are specified then the current state is returned
        """
        self._io_ad_config.value = states

    @property
    def _pin_states(self):
        """
        Contains data relating to zero or more pins.
        Structured as a variable length list of up to 19 Pin
        Number / Value pairs.
        """
        return self._io_pin_data.value

    @_pin_states.setter
    def _pin_states(self, pin_value_pairs):
        """
        Contains data relating to zero or more pins.
        Structured as a variable length list of up to 19 Pin
        Number / Value pairs.
        """
        self._io_pin_data.value = pin_value_pairs

    @property
    def pin_values(self):
        """
        Get the values of all the pins that are set as inputs

        :return: Dictionary (keys are pins)
        """
        return_dict = {}
        values = self._io_pin_data.value
        for i in range(0, len(values), 2):
            return_dict[str(int(values[i]))] = int(values[i + 1])
        return return_dict

    @property
    def _pin_pwm_control(self):
        """
        Write only method to set the PWM control data

        :param pin: pin number [range 0-19]
        :param value: Value is in the range 0 to 1024, per the current DAL API
            (e.g. setAnalogValue). 0 means OFF.
        :param period: Period is in microseconds and is an unsigned integer
        :return:
        """
        return self._io_pin_pwm.value

    @_pin_pwm_control.setter
    def _pin_pwm_control(self, data):
        """
        Write only method to set the PWM control data

        :param pin: pin number [range 0-19]
        :param value: Value is in the range 0 to 1024, per the current DAL API
            (e.g. setAnalogValue). 0 means OFF.
        :param period: Period is in microseconds and is an unsigned integer
        :return:
        """
        pin = data[0]
        value = data[1]
        period = data[2]
        byte_value = tools.int_to_uint16(value)
        byte_period = tools.int_to_uint32(period)
        self._io_pin_pwm.value = [pin,
                                  byte_value[0],
                                  byte_value[1],
                                  byte_period[0],
                                  byte_period[1],
                                  byte_period[2],
                                  byte_period[3]
                                  ]

    @property
    def uart(self):
        """
        Write string to micro:bit UART service. To read from UART, use

        :return:
        """
        pass

    @uart.setter
    def uart(self, txt):
        data = []
        text = ''
        if len(txt) > 20:
            text = txt[:19]
        else:
            text = txt
        for letter in text:
            data.append(ord(letter))
        self._uart_rx.value = data

    def subscribe_uart(self, user_callback):
        """
        Execute user_callback on data being received on UART service

        :param user_callback:
        :return:
        """
        self.uart_tx_cb = user_callback
        self._uart_tx.add_characteristic_cb(self._uart_read)
        self._uart_tx.start_notify()

    def _uart_read(self, iface, changed_props, invalidated_props):
        if iface != constants.GATT_CHRC_IFACE:
            return
        if 'Value' in changed_props:
            self.uart_tx_cb(''.join([str(v) for v in changed_props['Value']]))

    @property
    def on_disconnect(self):
        """Add callback for on_disconnect action"""
        return self.ubit.dongle.on_disconnect

    @on_disconnect.setter
    def on_disconnect(self, callback):
        self.ubit.dongle.on_disconnect = callback

    def run_async(self):
        """
        Puts the code into asynchronous mode

        :return:
        """
        self.ubit.run()

    def quit_async(self):
        """
        Stops asynchronous mode

        :return:
        """
        self.ubit.quit()


class MIpower(Microbit):
    """
    Initialization of an instance of a remote micro:bit
    with a MI:power board attached

    :param device_addr: Connect to a BLE device with this address
    :param adapter_addr: Use the adapter with this address
    """
    def __init__(self, device_addr, adapter_addr=None,
                 accelerometer_service=True,
                 button_service=True,
                 led_service=True,
                 magnetometer_service=False,
                 pin_service=False,
                 temperature_service=True
                 ):

        Microbit.__init__(self, device_addr, adapter_addr,
                          accelerometer_service=accelerometer_service,
                          button_service=button_service,
                          led_service=led_service,
                          magnetometer_service=magnetometer_service,
                          pin_service=pin_service,
                          temperature_service=temperature_service)

    def beep(self, duration=1):
        """
        If a buzzer is attached to pin 0 then a beep will be played

        :param duration: time in seconds
        """
        self._pin_pwm_control = [0, 512, 2094]
        sleep(duration)
        self._pin_pwm_control = [0, 0, 0]


class BitBot:
    """
    Class to simplify interacting with a microbit attached to a bit:bot
    over Bluetooth Low Energy
    The bit:bot is a micro:bit robot available from 4tronix.co.uk
    """
    def __init__(self, device_addr, adapter_addr=None,
                 accelerometer_service=False,
                 button_service=True,
                 led_service=True,
                 magnetometer_service=False,
                 pin_service=True,
                 temperature_service=False):
        """
        Initialization of an instance of a remote bit:bot

        :param name: Will look for a BLE device with this string in its name
        :param device_addr: Will look for a BLE device with this address
        """
        self._pins_configured = False
        self.ubit = Microbit(device_addr, adapter_addr,
                             accelerometer_service=accelerometer_service,
                             button_service=button_service,
                             led_service=led_service,
                             magnetometer_service=magnetometer_service,
                             pin_service=pin_service,
                             temperature_service=temperature_service)

    def __enter__(self):
        return self

    def __del__(self):
        self.clean_up()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clean_up()

    def clean_up(self):
        """
        Stop bitbot and turn buzzer is off

        :return:
        """
        if self.connected:
            self.stop()
            self.buzzer_off()

    def connect(self):
        """
        Connect to the bit:bot
        """
        self.ubit.connect()
        self._config_pins()

    def disconnect(self):
        """
        Disconnect from the bit:bot
        """
        self.clean_up()
        self.ubit.disconnect()

    @property
    def connected(self):
        """
        Returns true if bit:bot is connected
        """
        return self.ubit.connected

    def stop(self):
        """
        Stop both wheels of the bit:bot
        """
        self.ubit._pin_states = [0x01, 0x00,
                                 0x0C, 0x00,
                                 0x00, 0x00,
                                 0x08, 0x00]

    def spin_right(self):
        """
        Spin right wheel forward and left wheel backwards so bit:bot spins
        """
        self.ubit._pin_states = [0x01, 0x01,
                                 0x0C, 0x00,
                                 0x00, 0x00,
                                 0x08, 0x01]

    def spin_left(self):
        """
        Spin left wheel forward and right wheel backwards so bit:bot spins
        """
        self.ubit._pin_states = [0x01, 0x00,
                                 0x0C, 0x01,
                                 0x00, 0x01,
                                 0x08, 0x00]

    def forward(self):
        """
        Spin both wheels forward
        """
        self.ubit._pin_states = [0x01, 0x01,
                                 0x0C, 0x00,
                                 0x00, 0x01,
                                 0x08, 0x00]

    def reverse(self):
        """
        Spin both wheels backwards
        """
        self.ubit._pin_states = [0x01, 0x00,
                                 0x0C, 0x01,
                                 0x00, 0x00,
                                 0x08, 0x01]

    def _left_motor(self, pwm_value, reverse=False):
        if not reverse:
            self.ubit._pin_states = [0x08, 0x00]
        else:
            self.ubit._pin_states = [0x08, 0x01]
        self.ubit._pin_pwm_control = [0, pwm_value, 20000]

    def _right_motor(self, pwm_value, reverse=False):
        if not reverse:
            self.ubit._pin_states = [0x0C, 0x00]
        else:
            self.ubit._pin_states = [0x0C, 0x01]
        self.ubit._pin_pwm_control = [1, pwm_value, 20000]

    def _update_motors(self, left_val, right_val,
                       left_rev, right_rev,
                       pwm_period=20000):
        period = tools.int_to_uint32(pwm_period)
        left_pwm = tools.int_to_uint16(left_val)
        right_pwm = tools.int_to_uint16(right_val)
        self.ubit._io_pin_pwm.value = [0,
                                       left_pwm[0],
                                       left_pwm[1],
                                       period[0],
                                       period[1],
                                       period[2],
                                       period[3],
                                       1,
                                       right_pwm[0],
                                       right_pwm[1],
                                       period[0],
                                       period[1],
                                       period[2],
                                       period[3]
                                       ]
        self.ubit._pin_states = [0x08, left_rev, 0x0C, right_rev]

    def drive(self, left=100, right=100):
        """
        Set the drive power of both wheels at same time

        :param left: percentage of power (negative numbers are reverse)
        :param right: percentage of power (negative numbers are reverse)
        """
        left_direction = 0
        left_motor = self._percentage_to_pwm(left)

        right_direction = 0
        right_motor = self._percentage_to_pwm(right)

        if left < 0:
            left_direction = 1

        if right < 0:
            right_direction = 1

        self._update_motors(left_motor, right_motor,
                            left_direction, right_direction)

    def _percentage_to_pwm(self, percent):
        if percent < 0:
            percent += 100
        return int(10.23 * percent)

    def buzzer_on(self):
        """
        Play the buzzer
        """
        self.ubit._pin_states = [0x0E, 0x01]

    def buzzer_off(self):
        """
        Stop the buzzer
        """
        self.ubit._pin_states = [0x0E, 0x00]

    @property
    def left_line_senor(self):
        """
        Value ofthe left line sensor

        :return: False = No line  True = Line
        """
        return bool(self._get_pin_value(11))

    @property
    def right_line_sensor(self):
        """
        Value of the right line sensor

        :return: False = No line  True = Line
        """
        return bool(self._get_pin_value(5))

    @property
    def line_sensors(self):
        """
        Get the value of both line sensors

        :return: (left, right)
        """
        if not self._pins_configured:
            self._config_pins()
            self._pins_configured = True
        pins = self.ubit._pin_states
        return bool(pins[3]), bool(pins[5])

    @property
    def left_light_sensor(self):
        """
        Get the value of the left light sensor
        """
        self.ubit._pin_states = [0x10, 0x00]
        return int(self._get_pin_value(2))

    @property
    def right_light_sensor(self):
        """
        Get the value of the left light sensor
        """
        self.ubit._pin_states = [0x10, 0x01]
        return int(self._get_pin_value(2))

    def _config_pins(self):
        self.ubit._pin_config = [0x24, 0x08, 0x00, 0x00]
        self.ubit._pin_ad_config = [0x04, 0x00, 0x00, 0x00]

    def _build_pin_value_pairs(self, pin_states):
        val_dict = {}
        if len(pin_states) > 0:
            val_dict = dict(zip(pin_states[::2], pin_states[1::2]))
        return val_dict

    def _get_pin_value(self, pin):
        if not self._pins_configured:
            self._config_pins()
            self._pins_configured = True
        pin_value_pairs = self._build_pin_value_pairs(self.ubit._pin_states)
        if pin in pin_value_pairs.keys():
            return pin_value_pairs[pin]
        else:
            return None


class BitCommander:
    """
    Class to simplify interacting with a microbit attached to a bit:commander
    over Bluetooth Low Energy
    The bit:commander is a micro:bit controller available from 4tronix.co.uk
    """
    def __init__(self, device_addr, adapter_addr=None,
                 accelerometer_service=True,
                 button_service=True,
                 led_service=True,
                 magnetometer_service=False,
                 pin_service=False,
                 temperature_service=True):
        """
        Initialization of an instance of a remote bit:bot

        :param name: Will look for a BLE device with this string in its name
        :param device_addr: Will look for a BLE device with this address
        """
        self._pins_configured = False
        self.ubit = Microbit(device_addr, adapter_addr,
                             accelerometer_service=accelerometer_service,
                             button_service=button_service,
                             led_service=led_service,
                             magnetometer_service=magnetometer_service,
                             pin_service=pin_service,
                             temperature_service=temperature_service)

    def __enter__(self):
        return self

    def __del__(self):
        self.clean_up()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clean_up()

    def clean_up(self):
        """Put BitCommander back to known state"""
        if self.connected:
            pass

    def connect(self):
        """
        Connect to the bit:bot
        """
        self.ubit.connect()
        self._config_pins()

    def disconnect(self):
        """
        Disconnect from the bit:bot
        """
        self.clean_up()
        self.ubit.disconnect()

    @property
    def connected(self):
        """
        Returns true if bit:bot is connected
        """
        return self.ubit.connected

    def _config_pins(self):
        # Buttons
        self.ubit.set_pin(12, True, False)
        self.ubit.set_pin(14, True, False)
        self.ubit.set_pin(15, True, False)
        self.ubit.set_pin(16, True, False)
        # Dial
        self.ubit.set_pin(2, True, True)
        # Joy sticks
        self.ubit.set_pin(8, True, False)
        self.ubit.set_pin(1, True, True)
        self.ubit.set_pin(0, True, True)
        self._pins_configured = True

    def subscribe_pins(self, user_cb):
        """
        Get notification when pins configured as input values change

        :param user_cb: User callback to call when change happens
        :return:
        """
        self.ubit.subscribe_pins(user_cb)

    @property
    def joystick(self):
        """
        The value of the joystick as a list of x, y and z
        x and y are analogue values in the range -255 to 255
        z is a button with a binary value of 1 for pressed
        :return: x, y, z
        """
        values = self.ubit.pin_values
        if '1' in values:
            x, y, z = (values['1'],  # pylint: disable=invalid-name
                       values['0'], values['8'])
        return x, y, z

    @property
    def dial(self):
        """
        The value of the dial.

        :return: integer in range 0 - 255
        """
        values = self.ubit.pin_values
        return values['2']

    def _read_button(self, pin):
        values = self.ubit.pin_values
        if pin in values:
            return values[pin]
        else:
            return None

    @property
    def button_a(self):
        """
        The state of button a of bit:commander

        :return: when pressed returns 1
        """
        return self._read_button('12')

    @property
    def button_b(self):
        """
        The state of button b of bit:commander

        :return: when pressed returns 1
        """
        return self._read_button('14')

    @property
    def button_c(self):
        """
        The state of button c of bit:commander

        :return: when pressed returns 1
        """
        return self._read_button('16')

    @property
    def button_d(self):
        """
        The state of button d of bit:commander

        :return: when pressed returns 1
        """
        return self._read_button('15')

    def run_async(self):
        """
        Puts the code into asynchronous mode

        :return:
        """
        self.ubit.run_async()

    def quit_async(self):
        """
        Stops asynchronose mode

        :return:
        """
        self.ubit.quit_async()



"""Classes required to create a Bluetooth Peripheral."""

# python-bluezero imports
from bluezero import adapter
from bluezero import advertisement
from bluezero import async_tools
from bluezero import localGATT
from bluezero import GATT
from bluezero import tools


logger = tools.create_module_logger(__name__)


class Peripheral:
    """Create a Bluetooth BLE Peripheral"""
    def __init__(self, adapter_address, local_name=None, appearance=None):
        self.app = localGATT.Application()
        self.srv_mng = GATT.GattManager(adapter_address)
        self.services = []
        self.characteristics = []
        self.descriptors = []
        self.primary_services = []
        self.dongle = adapter.Adapter(adapter_address)
        self.local_name = local_name
        self.appearance = appearance
        self.advert = advertisement.Advertisement(1, 'peripheral')
        self.ad_manager = advertisement.AdvertisingManager(adapter_address)
        self.mainloop = async_tools.EventLoop()

    def add_service(self, srv_id, uuid, primary):
        """
        Add the service information required

        :param srv_id: integer between 0 & 9999 as unique reference
        :param uuid: The Bluetooth uuid number for this service
        :param primary: boolean for if this service should be advertised

        """
        self.services.append(localGATT.Service(srv_id, uuid, primary))
        if primary:
            self.primary_services.append(uuid)

    def add_characteristic(self, srv_id, chr_id, uuid, value,
                           notifying, flags,
                           read_callback=None, write_callback=None,
                           notify_callback=None):
        """
        Add information for characteristic.

        :param srv_id: integer of parent service that was added
        :param chr_id: integer between 0 & 9999 as unique reference
        :param uuid: The Bluetooth uuid number for this characteristic
        :param value: Initial value. list of integers in little endian format
        :param notifying: Boolean representing initial state of notifications
        :param flags: Defines how the characteristic value can be used. See
            Core spec "Table 3.5: Characteristic Properties bit field", and
            "Table 3.8: Characteristic Extended. Properties bit field".
            Allowed values:

            - "broadcast"
            - "read"
            - "write-without-response"
            - "write"
            - "notify"
            - "indicate"
            - "authenticated-signed-writes"
            - "extended-properties"
            - "reliable-write"
            - "writable-auxiliaries"
            - "encrypt-read"
            - "encrypt-write"
            - "encrypt-authenticated-read"
            - "encrypt-authenticated-write"
            - "secure-read" (Server only)
            - "secure-write" (Server only)
            - "authorize"

        :param read_callback: function to be called when read_value is called
            by client. function should return python list of integers
            representing new value of characteristic
        :param write_callback: function to be called when write_value is called
            by client. Function should have two parameters value and options.
            value is python list of integers with new value of characteristic.
        :param notify_callback: function to be called when notify_start or
            notify_stop is called by client. Function should have two
            parameters notifying and characteristic. The `characteristic`
            is the instantiation of a localGAT.Characteristic class

        """
        self.characteristics.append(localGATT.Characteristic(
            srv_id, chr_id, uuid, value, notifying, flags,
            read_callback, write_callback, notify_callback
        ))

    def add_descriptor(self, srv_id, chr_id, dsc_id, uuid, value, flags):
        """
        Add information for the GATT descriptor.

        :param srv_id: integer of parent service that was added
        :param chr_id: integer of parent characteristic that was added
        :param dsc_id: integer between 0 & 9999 as unique reference
        :param uuid: The Bluetooth uuid number for this characteristic
        :param value: Initial value. list of integers in little endian format
        :param flags: Defines how the descriptor value can be used.
            Possible values:

                - "read"
                - "write"
                - "encrypt-read"
                - "encrypt-write"
                - "encrypt-authenticated-read"
                - "encrypt-authenticated-write"
                - "secure-read" (Server Only)
                - "secure-write" (Server Only)
                - "authorize"

        """
        self.descriptors.append(localGATT.Descriptor(
            srv_id, chr_id, dsc_id, uuid, value, flags
        ))

    def _create_advertisement(self):
        self.advert.service_UUIDs = self.primary_services
        if self.local_name:
            self.advert.local_name = self.local_name
        if self.appearance:
            self.advert.appearance = self.appearance

    def publish(self):
        """Create advertisement and make peripheral visible"""
        for service in self.services:
            self.app.add_managed_object(service)
        for chars in self.characteristics:
            self.app.add_managed_object(chars)
        for desc in self.descriptors:
            self.app.add_managed_object(desc)
        self._create_advertisement()
        if not self.dongle.powered:
            self.dongle.powered = True
        self.srv_mng.register_application(self.app, {})
        self.ad_manager.register_advertisement(self.advert, {})

        try:
            self.mainloop.run()
        except KeyboardInterrupt:
            self.mainloop.quit()
            self.ad_manager.unregister_advertisement(self.advert)

    @property
    def on_connect(self):
        """
        Callback for when a device connects to the peripheral.
        Callback can accept 0, 1, or 2 positional arguments
        1: a device.Device instance of the connected target
        2: the local adapter address followed by the remote address
        """
        return self.dongle.on_connect

    @on_connect.setter
    def on_connect(self, callback):
        self.dongle.on_connect = callback

    @property
    def on_disconnect(self):
        """
        Callback for when a device disconnects from the peripheral.
        Callback can accept 0, 1, or 2 positional arguments
        1: a device.Device instance of the disconnected target
        2: the local adapter address followed by the remote address
        """
        return self.dongle.on_disconnect

    @on_disconnect.setter
    def on_disconnect(self, callback):
        self.dongle.on_disconnect = callback



"""Automated testing of GATT functionality using unittest.mock."""
import sys
import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants

adapter_props = tests.obj_data.full_ubits


def mock_get(iface, prop):
    if iface == 'org.bluez.Adapter1':
        return tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop]
    elif iface == 'org.bluez.Device1':
        return tests.obj_data.full_ubits['/org/bluez/hci0/dev_E4_43_33_7E_54_1C'][iface][prop]
    else:
        return tests.obj_data.full_ubits['/org/bluez/hci0/dev_E4_43_33_7E_54_1C/service002a'][iface][prop]


def mock_set(iface, prop, value):
    tests.obj_data.full_ubits['/org/bluez/hci0/dev_E4_43_33_7E_54_1C/service002a'][iface][prop] = value


class TestBluezeroCentral(unittest.TestCase):
    """Test class to exercise Central Features."""

    def setUp(self):
        """Initialise the class for the tests."""
        self.dbus_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.exceptions': self.dbus_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
        }
        self.dbus_mock.Interface.return_value.GetManagedObjects.return_value = tests.obj_data.full_ubits
        self.dbus_mock.Interface.return_value.Get = mock_get
        self.dbus_mock.Interface.return_value.Set = mock_set
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import central
        self.module_under_test = central
        self.path = '/org/bluez/hci0/dev_E4_43_33_7E_54_1C/service002a'
        self.adapter_path = '/org/bluez/hci0'
        self.dev_name = 'BBC micro:bit [zezet]'
        self.adapter_addr = '00:00:00:00:5A:AD'
        self.device_addr = 'E4:43:33:7E:54:1C'
        self.service_uuid = 'e95dd91d-251d-470a-a062-fa1922dfa9a8'

    def tearDown(self):
        self.module_patcher.stop()

    def test_service_uuid(self):
        """Test the central instantiation."""
        # Invoke the bluez GATT library to access the mock GATT service
        test_central = self.module_under_test.Central(adapter_addr=self.adapter_addr,
                                                      device_addr=self.device_addr)

        # Test for the UUID
        self.assertEqual(test_central.connected, True)




"""Class and methods that represent a Bluetooth Adapter."""

# D-Bus imports
import dbus
import dbus.mainloop.glib

# python-bluezero imports
from bluezero import constants
from bluezero import dbus_tools
from bluezero import async_tools
from bluezero import device
from bluezero import tools


logger = tools.create_module_logger(__name__)

dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)


class AdapterError(Exception):
    """Custom exception for missing Bluetooth adapter"""
    pass


def list_adapters():
    """Return list of adapters address available on system."""
    return [dongle.address for dongle in Adapter.available()]


class Adapter:
    """Bluetooth Adapter Class.

    This class instantiates an object that interacts with the physical
    Bluetooth device.

    :Example:

    >>> from bluezero import adapter
    >>> dongle = adapter.Adapter()
    >>> dongle.powered = True

    """

    @staticmethod
    def available():
        """
        A generator yielding an Adapter object for every attached adapter.
        """
        mng_objs = dbus_tools.get_managed_objects()
        found = False
        for obj in mng_objs.values():
            adapter = obj.get(constants.ADAPTER_INTERFACE, None)
            if adapter:
                found = True
                yield Adapter(adapter['Address'])
        if not found:
            raise AdapterError('No Bluetooth adapter found')

    def __init__(self, adapter_addr=None):
        """Default initialiser.

        Creates the interface to the local Bluetooth adapter device.
        If address is not given then first device is list is used.

        :param adapter_addr: Address of Bluetooth adapter to use.
        """
        self.bus = dbus.SystemBus()

        if adapter_addr is None:
            adapters = list_adapters()
            if len(adapters) > 0:
                adapter_addr = adapters[0]

        self.path = dbus_tools.get_dbus_path(adapter=adapter_addr)
        self.adapter_object = self.bus.get_object(
            constants.BLUEZ_SERVICE_NAME,
            self.path)
        self.adapter_methods = dbus.Interface(self.adapter_object,
                                              constants.ADAPTER_INTERFACE)

        self.adapter_props = dbus.Interface(self.adapter_object,
                                            dbus.PROPERTIES_IFACE)

        self._nearby_timeout = 10
        self._nearby_count = 0
        self.mainloop = async_tools.EventLoop()

        self.on_disconnect = None
        self.on_connect = None
        self.on_device_found = None
        self.bus.add_signal_receiver(self._interfaces_added,
                                     dbus_interface=constants.DBUS_OM_IFACE,
                                     signal_name='InterfacesAdded')

        self.bus.add_signal_receiver(self._interfaces_removed,
                                     dbus_interface=constants.DBUS_OM_IFACE,
                                     signal_name='InterfacesRemoved')

        self.bus.add_signal_receiver(self._properties_changed,
                                     dbus_interface=dbus.PROPERTIES_IFACE,
                                     signal_name='PropertiesChanged',
                                     arg0=constants.DEVICE_INTERFACE,
                                     path_keyword='path')

    @property
    def address(self):
        """Return the adapter MAC address."""
        return self.adapter_props.Get(constants.ADAPTER_INTERFACE, 'Address')

    @property
    def name(self):
        """Return the adapter name."""
        return self.adapter_props.Get(constants.ADAPTER_INTERFACE, 'Name')

    @property
    def bt_class(self):
        """Return the Bluetooth class of device."""
        return self.adapter_props.Get(constants.ADAPTER_INTERFACE, 'Class')

    @property
    def alias(self):
        """Return the adapter alias.

        :param new_alias: the new alias of the adapter.
        """
        return self.adapter_props.Get(
            constants.ADAPTER_INTERFACE, 'Alias')

    @alias.setter
    def alias(self, new_alias):
        self.adapter_props.Set(
            constants.ADAPTER_INTERFACE, 'Alias', new_alias)

    def get_all(self):
        """Return dictionary of all the Adapter attributes."""
        return self.adapter_props.GetAll(constants.ADAPTER_INTERFACE)

    @property
    def powered(self):
        """power state of the Adapter.

        :param new_state: boolean.
        """
        return self.adapter_props.Get(
            constants.ADAPTER_INTERFACE, 'Powered')

    @powered.setter
    def powered(self, new_state):
        self.adapter_props.Set(
            constants.ADAPTER_INTERFACE, 'Powered', new_state)

    @property
    def pairable(self):
        """pairable state of the Adapter.

        :param new_state: boolean.
        """
        return self.adapter_props.Get(
            constants.ADAPTER_INTERFACE, 'Pairable')

    @pairable.setter
    def pairable(self, new_state):
        self.adapter_props.Set(
            constants.ADAPTER_INTERFACE, 'Pairable', new_state)

    @property
    def pairabletimeout(self):
        """The pairable timeout of the Adapter."""
        return self.adapter_props.Get(constants.ADAPTER_INTERFACE,
                                      'PairableTimeout')

    @pairabletimeout.setter
    def pairabletimeout(self, new_timeout):
        self.adapter_props.Set(constants.ADAPTER_INTERFACE,
                               'PairableTimeout', new_timeout)

    @property
    def discoverable(self):
        """Discoverable state of the Adapter."""
        return self.adapter_props.Get(
            constants.ADAPTER_INTERFACE, 'Discoverable')

    @discoverable.setter
    def discoverable(self, new_state):
        self.adapter_props.Set(constants.ADAPTER_INTERFACE,
                               'Discoverable', new_state)

    @property
    def discoverabletimeout(self):
        """Discoverable timeout of the Adapter."""
        return self.adapter_props.Get(constants.ADAPTER_INTERFACE,
                                      'DiscoverableTimeout')

    @discoverabletimeout.setter
    def discoverabletimeout(self, new_timeout):
        self.adapter_props.Set(constants.ADAPTER_INTERFACE,
                               'DiscoverableTimeout', new_timeout)

    @property
    def discovering(self):
        """Return whether the adapter is discovering."""
        return self.adapter_props.Get(
            constants.ADAPTER_INTERFACE, 'Discovering')

    def _discovering_timeout(self):
        """Test to see if discovering should stop."""
        self._nearby_count += 1
        if self._nearby_count > self._nearby_timeout:
            self.stop_discovery()
            self.mainloop.quit()
            return False
        return True

    @property
    def uuids(self):
        """List of 128-bit UUIDs that represent available remote services."""
        return self.adapter_props.Get(
            constants.ADAPTER_INTERFACE, 'UUIDs')

    def nearby_discovery(self, timeout=10):
        """Start discovery of nearby Bluetooth devices."""
        self._nearby_timeout = timeout
        self._nearby_count = 0

        # GLib.timeout_add(1000, self._discovering_timeout)
        async_tools.add_timer_ms(1000, self._discovering_timeout)
        self.adapter_methods.StartDiscovery()
        self.mainloop.run()

    def show_duplicates(self):
        """
        Show every advertisement from a device during
        Device Discovery if it contains ManufacturerData and/or
        ServiceData irrespective of whether they have been
        discovered previously
        """
        self.adapter_methods.SetDiscoveryFilter({'DuplicateData': True})

    def hide_duplicates(self):
        """
        Hide advertisements from a device during
        Device Discovery if it contains information already discovered
        """
        self.adapter_methods.SetDiscoveryFilter({'DuplicateData': False})

    def start_discovery(self):
        """
        Start discovery of nearby Bluetooth devices.

        :return: True on success otherwise False
        """
        self.adapter_methods.StartDiscovery()

    def stop_discovery(self):
        """Stop scanning of nearby Bluetooth devices."""
        self.adapter_methods.StopDiscovery()

    def remove_device(self, device_path):
        """Removes device at the given D-Bus path"""
        self.adapter_methods.RemoveDevice(device_path)

    def run(self):
        """Start the EventLoop for async operations"""
        self.mainloop.run()

    def quit(self):
        """Stop the EventLoop for async operations"""
        self.mainloop.quit()

    def _properties_changed(self, interface, changed, invalidated, path):
        """
        Handle DBus PropertiesChanged signal and
        call appropriate user callback
        """
        device_address = dbus_tools.get_device_address_from_dbus_path(path)
        adapter_addr = dbus_tools.get_adapter_address_from_dbus_path(path)
        if 'Connected' in changed:
            if all((changed['Connected'],
                    self.address == adapter_addr,
                    self.on_connect)):
                if tools.get_fn_parameters(self.on_connect) == 0:
                    self.on_connect()
                elif tools.get_fn_parameters(self.on_connect) == 1:
                    new_dev = device.Device(
                        adapter_addr=self.address,
                        device_addr=device_address)
                    self.on_connect(new_dev)
                elif tools.get_fn_parameters(self.on_connect) == 2:
                    self.on_connect(adapter_addr, device_address)
            elif not changed['Connected'] and self.on_disconnect:
                if tools.get_fn_parameters(self.on_disconnect) == 0:
                    logger.warning("using deprecated version of disconnect "
                                   "callback, move to on_disconnect(dev) "
                                   "with device parameter")
                    self.on_disconnect()
                elif tools.get_fn_parameters(self.on_disconnect) == 1:
                    new_dev = device.Device(
                        adapter_addr=self.address,
                        device_addr=device_address)
                    self.on_disconnect(new_dev)
                elif tools.get_fn_parameters(self.on_disconnect) == 2:
                    self.on_disconnect(self.address, device_address)

    def _interfaces_added(self, path, device_info):
        """
        Handle DBus InterfacesAdded signal and
        call appropriate user callback
        """
        dev_iface = constants.DEVICE_INTERFACE
        if constants.DEVICE_INTERFACE in device_info:
            dev_addr = device_info[dev_iface].get('Address')
            dev_connected = device_info[dev_iface].get('Connected')
            if self.on_device_found and dev_addr:
                new_dev = device.Device(
                    adapter_addr=self.address,
                    device_addr=dev_addr)
                self.on_device_found(new_dev)
            if all((self.on_connect, dev_connected, dev_addr)):
                new_dev = device.Device(
                    adapter_addr=self.address,
                    device_addr=dev_addr)
                self.on_connect(new_dev)

    def _interfaces_removed(self, path, device_info):
        """
        Handle DBus InterfacesRemoved signal and
        call appropriate user callback
        """



import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants

adapter_props = tests.obj_data.full_ubits


def mock_get(iface, prop):
    return tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop]


def mock_set(iface, prop, value):
    tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop] = value


class TestBluezeroEddystone(unittest.TestCase):

    def setUp(self):
        """
        Patch the DBus module
        :return:
        """
        self.dbus_mock = MagicMock()
        self.dbus_exception_mock = MagicMock()
        self.dbus_service_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.exceptions': self.dbus_exception_mock,
            'dbus.service': self.dbus_service_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
        }
        self.dbus_mock.Interface.return_value.GetManagedObjects.return_value = tests.obj_data.full_ubits
        self.dbus_mock.Interface.return_value.Get = mock_get
        self.dbus_mock.Interface.return_value.Set = mock_set
        self.dbus_mock.SystemBus = MagicMock()
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import eddystone_beacon
        self.module_under_test = eddystone_beacon

    def tearDown(self):
        self.module_patcher.stop()

    def test_long_url(self):
        self.assertRaises(Exception,
                          self.module_under_test.EddystoneURL,
                          'https://ukbaz.github.io/howto/beacon_scan_cmd_line.html')

    def test_load(self):
        self.module_under_test.EddystoneURL('http://camjam.me')



"""Classes required to create a Bluetooth Peripheral."""

from inspect import signature

# D-Bus imports
import dbus
import dbus.exceptions
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop

# python-bluezero imports
from bluezero import async_tools
from bluezero import constants
from bluezero import dbus_tools
from bluezero import tools

DBusGMainLoop(set_as_default=True)

logger = tools.create_module_logger(__name__)


########################################
# Exception classes
#######################################
class InvalidArgsException(dbus.exceptions.DBusException):
    """This is a D-Bus exception class for Invalid Arguments.

    All this class does is set the internal variable ``_dbus_error_name`` to
    the object path for D-Bus Invalid Argument Exceptions.

    """

    _dbus_error_name = 'org.freedesktop.DBus.Error.InvalidArgs'


class Application(dbus.service.Object):
    """Bluez Application Class.

    This is the parent class for a python Bluez application.

    """
    def __init__(self):
        """Default initialiser.

        """
        # Initialise the D-Bus path and register it
        self.bus = dbus.SystemBus()
        self.path = dbus.ObjectPath(constants.BLUEZERO_DBUS_OBJECT)
        dbus.service.Object.__init__(self, self.bus, self.path)

        # Objects to be associated with this service
        self.managed_objs = []
        self.eventloop = async_tools.EventLoop()

    @dbus.service.method(constants.DBUS_OM_IFACE,
                         out_signature='a{oa{sa{sv}}}')
    def GetManagedObjects(self):  # pylint: disable=invalid-name
        """Get all objects that are managed by the application.

        Return type is a dictionary whose keys are each registered object and
        values the properties of the given object.
        """
        response = {}

        for obj in self.managed_objs:
            for iface in obj.props.keys():
                response[obj.get_path()] = {iface: obj.GetAll(iface)}

        return response

    def add_managed_object(self, service_obj):
        """Add a service to the list of services offered by the Application.

        :param service_obj: Python object of dbus path to be managed
        """
        self.managed_objs.append(service_obj)

    def get_path(self):
        """Return the DBus object path"""
        return dbus.ObjectPath(self.path)

    def start(self):
        """Start event loop"""
        self.eventloop.run()

    def stop(self):
        """Stop event loop"""
        self.eventloop.quit()


class Service(dbus.service.Object):
    """Bluez Service Class.

    This class represents a BLE Service.

    :Example:

    >>> your_service = localGATT.Service(service_id, UUID, primary)

    """

    PATH_BASE = f'{constants.BLUEZERO_DBUS_OBJECT}/service'

    def __init__(self, service_id, uuid, primary):
        """Default initialiser.

        1. Registers the service on the D-Bus.
        2. Sets up the service UUID and primary flags.

        :param service_id:
        :param uuid: service BLE UUID
        :param primary: whether or not the service is a primary service
        """
        # Setup D-Bus object paths and register service
        self.path = self.PATH_BASE + str('{0:04d}'.format(service_id))
        self.bus = dbus.SystemBus()
        self.interface = constants.GATT_SERVICE_IFACE
        dbus.service.Object.__init__(self, self.bus, self.path)
        self.props = {
            constants.GATT_SERVICE_IFACE: {
                'UUID': uuid,
                'Primary': primary}
        }

    def get_path(self):
        """Return the DBus object path"""
        return dbus.ObjectPath(self.path)

    @dbus.service.method(constants.DBUS_PROP_IFACE,
                         in_signature='s',
                         out_signature='a{sv}')
    def GetAll(self, interface_name):  # pylint: disable=invalid-name
        """Return the service properties.

        This method is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``

        :param interface_name: interface to get the properties of.

        The interface must be ``org.bluez.GattService1`` otherwise an
        exception is raised.
        """
        if interface_name != constants.GATT_SERVICE_IFACE:
            raise InvalidArgsException()

        try:
            return self.props[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such interface ' + interface_name,
                name=interface_name + '.UnknownInterface')

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ss', out_signature='v')
    def Get(self,  # pylint: disable=invalid-name
            interface_name, property_name):
        """
        DBus API for getting a property value

        :param interface_name:
        :param property_name:
        :return:
        """

        if interface_name != constants.GATT_SERVICE_IFACE:
            raise InvalidArgsException()

        try:
            return self.GetAll(interface_name)[property_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=interface_name + '.UnknownProperty')

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ssv', out_signature='')
    def Set(self, interface_name,  # pylint: disable=invalid-name
            property_name, value):
        """Standard D-Bus API for setting a property value"""

        try:
            iface_props = self.props[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such interface ' + interface_name,
                name=self.interface + '.UnknownInterface')

        if property_name not in iface_props:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=self.interface + '.UnknownProperty')

        iface_props[property_name] = value

        self.PropertiesChanged(interface_name,
                               dbus.Dictionary({property_name: value},
                                               signature='sv'),
                               dbus.Array([], signature='s'))

    @dbus.service.signal(constants.DBUS_PROP_IFACE,
                         signature='sa{sv}as')
    def PropertiesChanged(self,  # pylint: disable=invalid-name
                          interface, changed, invalidated):
        """Emit a Properties Changed notification signal.

        This signal is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``.
        """
        logger.debug('Service properties changed: %s, %s, %s',
                     interface, changed, invalidated)


class Characteristic(dbus.service.Object):
    """
    Bluez Characteristic Class.

    This class represents a BLE Characteristic.
    """
    def __init__(self,
                 service_id,
                 characteristic_id,
                 uuid,
                 value,
                 notifying,
                 flags,
                 read_callback=None,
                 write_callback=None,
                 notify_callback=None):
        self.read_callback = read_callback
        self.write_callback = write_callback
        self.notify_callback = notify_callback

        # Setup D-Bus object paths and register service
        service_path = (f'{constants.BLUEZERO_DBUS_OBJECT}/'
                        f'service{service_id:04d}')
        self.path = f'{service_path}/char{characteristic_id:04d}'
        self.bus = dbus.SystemBus()
        dbus.service.Object.__init__(self, self.bus, self.path)
        self.props = {
            constants.GATT_CHRC_IFACE: {
                'UUID': uuid,
                'Service': dbus_tools.get_dbus_obj(service_path),
                'Value': dbus.Array(value, signature='y'),
                'Notifying': notifying,
                'Flags': flags}
        }

        for prop in self.props[constants.GATT_CHRC_IFACE].keys():
            self.Set(constants.GATT_CHRC_IFACE,
                     prop,
                     self.props[constants.GATT_CHRC_IFACE][prop])

    def get_path(self):
        """Return the DBus object path"""
        return dbus.ObjectPath(self.path)

    @property
    def is_notifying(self):
        """Get the current notify status"""
        return bool(self.Get(constants.GATT_CHRC_IFACE, 'Notifying'))

    def set_value(self, value):
        """
        Set the value of the characteristic. Will create notify event
        if notifications are enabled

        :param value: list of integers in little endian format
        """
        self.Set(constants.GATT_CHRC_IFACE, 'Value',
                 dbus.Array(value, signature='y'))

    @dbus.service.method(constants.DBUS_PROP_IFACE,
                         in_signature='s',
                         out_signature='a{sv}')
    def GetAll(self, interface_name):  # pylint: disable=invalid-name
        """Return the service properties.

        This method is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``

        :param interface_name: interface to get the properties of.

        The interface must be ``org.bluez.GattCharacteristic1`` otherwise an
        exception is raised.
        """
        if interface_name != constants.GATT_CHRC_IFACE:
            raise InvalidArgsException()

        try:
            return self.props[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such interface ' + interface_name,
                name=interface_name + '.UnknownInterface')

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ss', out_signature='v')
    def Get(self,  # pylint: disable=invalid-name
            interface_name, property_name):
        """DBus API for getting a property value.

        This method is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``

        :param interface_name: interface to get the properties of.
        :param property_name: request this property
        """

        if interface_name != constants.GATT_CHRC_IFACE:
            raise InvalidArgsException()

        try:
            return self.GetAll(interface_name)[property_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=interface_name + '.UnknownProperty')

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ssv', out_signature='')
    def Set(self, interface_name,   # pylint: disable=invalid-name
            property_name, value):
        """Standard D-Bus API for setting a property value"""

        if property_name not in self.props[constants.GATT_CHRC_IFACE]:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=constants.GATT_CHRC_IFACE + '.UnknownProperty')
        self.props[constants.GATT_CHRC_IFACE][property_name] = value
        return self.PropertiesChanged(interface_name,
                                      dbus.Dictionary({property_name: value},
                                                      signature='sv'),
                                      dbus.Array([], signature='s'))

    @dbus.service.signal(constants.DBUS_PROP_IFACE,
                         signature='sa{sv}as')
    def PropertiesChanged(self,  # pylint: disable=invalid-name
                          interface, changed, invalidated):
        """Emit a Properties Changed notification signal.

        This signal is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``.
        """
        logger.debug('Char Prop Changed')

    @dbus.service.method(constants.GATT_CHRC_IFACE,
                         in_signature='a{sv}', out_signature='ay')
    def ReadValue(self, options):  # pylint: disable=invalid-name
        """
        DBus method for getting the characteristic value

        :return: value
        """
        if self.read_callback:
            if len(signature(self.read_callback).parameters) == 1:
                value = self.read_callback(dbus_tools.dbus_to_python(options))
            else:
                value = self.read_callback()
            self.Set(constants.GATT_CHRC_IFACE, 'Value',
                     dbus.Array(value, signature='y'))
            logger.debug('ReadValue: %s', value)
        return self.GetAll(constants.GATT_CHRC_IFACE)['Value']

    @dbus.service.method(constants.GATT_CHRC_IFACE,
                         in_signature='aya{sv}', out_signature='')
    def WriteValue(self, value, options):  # pylint: disable=invalid-name
        """
        DBus method for setting the characteristic value

        :return: value
        """
        if self.write_callback:
            self.write_callback(dbus_tools.dbus_to_python(value),
                                dbus_tools.dbus_to_python(options))
        self.Set(constants.GATT_CHRC_IFACE, 'Value', value)

    @dbus.service.method(constants.GATT_CHRC_IFACE,
                         in_signature='', out_signature='')
    def StartNotify(self):  # pylint: disable=invalid-name
        """
        DBus method for enabling notifications of the characteristic value.

        :return: value
        """
        if self.props[constants.GATT_CHRC_IFACE]['Notifying'] is True:
            logger.info('Notifying already, nothing to do')
            return
        self.Set(constants.GATT_CHRC_IFACE,
                 'Notifying',
                 dbus.Boolean(True, variant_level=1))
        if self.notify_callback:
            self.notify_callback(True, self)

    @dbus.service.method(constants.GATT_CHRC_IFACE,
                         in_signature='', out_signature='')
    def StopNotify(self):  # pylint: disable=invalid-name
        """
        DBus method for disabling notifications of the characteristic value.

        :return: value
        """
        if self.props[constants.GATT_CHRC_IFACE]['Notifying'] is False:
            logger.info('Not Notifying, nothing to do')
            return

        self.Set(constants.GATT_CHRC_IFACE,
                 'Notifying',
                 dbus.Boolean(False, variant_level=1))
        if self.notify_callback:
            self.notify_callback(False, self)


class Descriptor(dbus.service.Object):
    """
    Bluez Descriptor Class.

    This class represents a BLE Descriptor.
    """
    def __init__(self,
                 service_id,
                 characteristic_id,
                 descriptor_id,
                 uuid,
                 value,
                 flags):
        # Setup D-Bus object paths and register service
        char_path = (f'{constants.BLUEZERO_DBUS_OBJECT}/'
                     f'service{service_id:04d}/char{characteristic_id:04d}')
        self.path = f'{char_path}/desc{descriptor_id:04d}'
        self.bus = dbus.SystemBus()
        dbus.service.Object.__init__(self, self.bus, self.path)
        self.props = {
            constants.GATT_DESC_IFACE: {
                'UUID': uuid,
                'Characteristic': dbus_tools.get_dbus_obj(char_path),
                'Value': value,
                'Flags': flags}
        }
        for prop in self.props[constants.GATT_DESC_IFACE].keys():
            self.Set(constants.GATT_DESC_IFACE,
                     prop,
                     self.props[constants.GATT_DESC_IFACE][prop])

    def get_path(self):
        """Return the DBus object path"""
        return dbus.ObjectPath(self.path)

    @dbus.service.method(constants.DBUS_PROP_IFACE,
                         in_signature='s',
                         out_signature='a{sv}')
    def GetAll(self, interface_name):  # pylint: disable=invalid-name
        """Return the descriptor properties.

        This method is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``

        :param interface: interface to get the properties of.

        The interface must be ``org.bluez.GattDescriptor1`` otherwise an
        exception is raised.
        """
        if interface_name != constants.GATT_DESC_IFACE:
            raise InvalidArgsException()

        try:
            return self.props[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such interface ' + interface_name,
                name=interface_name + '.UnknownInterface')

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ss', out_signature='v')
    def Get(self,   # pylint: disable=invalid-name
            interface_name, property_name):
        """DBus API for getting a property value.

        This method is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``

        :param interface_name: interface to get the properties of.
        :param property_name: request this property
        """

        if interface_name != constants.GATT_DESC_IFACE:
            raise InvalidArgsException()

        try:
            return self.GetAll(interface_name)[property_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=interface_name + '.UnknownProperty')

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ssv', out_signature='')
    def Set(self, interface_name,  # pylint: disable=invalid-name
            property_name, value):
        """Standard D-Bus API for setting a property value"""

        if property_name not in self.props[constants.GATT_DESC_IFACE]:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=constants.GATT_DESC_IFACE + '.UnknownProperty')

        self.props[interface_name][property_name] = value

        return self.PropertiesChanged(interface_name,
                                      dbus.Dictionary({property_name: value},
                                                      signature='sv'),
                                      dbus.Array([], signature='s'))

    @dbus.service.signal(constants.DBUS_PROP_IFACE,
                         signature='sa{sv}as')
    def PropertiesChanged(self,  # pylint: disable=invalid-name
                          interface, changed, invalidated):
        """Emit a Properties Changed notification signal.

        This signal is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``.
        """
        return 0

    @dbus.service.method(constants.GATT_DESC_IFACE,
                         in_signature='a{sv}', out_signature='ay')
    def ReadValue(self, options):  # pylint: disable=invalid-name
        """
        DBus method for getting the characteristic value

        :return: value
        """
        return self.GetAll(constants.GATT_DESC_IFACE)['Value']

    @dbus.service.method(constants.GATT_DESC_IFACE,
                         in_signature='aya{sv}', out_signature='')
    def WriteValue(self, value, options):  # pylint: disable=invalid-name
        """
        DBus method for setting the descriptor value

        :return:
        """
        return self.Set(constants.GATT_DESC_IFACE, 'Value', value)



import subprocess
import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants


class TestDbusModuleCalls(unittest.TestCase):
    """
    Testing things that use the Dbus module
    """
    experimental = True
    bluetooth_service_experimental = b'\xe2\x97\x8f bluetooth.service - Bluetooth service\n   Loaded: loaded (/lib/systemd/system/bluetooth.service; enabled; vendor preset: enabled)\n   Active: active (running) since Fri 2017-10-13 21:48:58 UTC; 1 day 23h ago\n     Docs: man:bluetoothd(8)\n Main PID: 530 (bluetoothd)\n   Status: "Running"\n   CGroup: /system.slice/bluetooth.service\n           \xe2\x94\x94\xe2\x94\x80530 /usr/lib/bluetooth/bluetoothd --experimental\n\nOct 13 21:48:58 RPi3 systemd[1]: Starting Bluetooth service...\nOct 13 21:48:58 RPi3 bluetoothd[530]: Bluetooth daemon 5.43\nOct 13 21:48:58 RPi3 systemd[1]: Started Bluetooth service.\nOct 13 21:48:58 RPi3 bluetoothd[530]: Bluetooth management interface 1.14 initialized\nOct 13 21:48:58 RPi3 bluetoothd[530]: Failed to obtain handles for "Service Changed" characteristic\nOct 13 21:48:58 RPi3 bluetoothd[530]: Endpoint registered: sender=:1.10 path=/MediaEndpoint/A2DPSource\nOct 13 21:48:58 RPi3 bluetoothd[530]: Endpoint registered: sender=:1.10 path=/MediaEndpoint/A2DPSink\n'
    bluetooth_service_normal = b'\xe2\x97\x8f bluetooth.service - Bluetooth service\n   Loaded: loaded (/lib/systemd/system/bluetooth.service; enabled; vendor preset: enabled)\n   Active: active (running) since Fri 2017-10-13 21:48:58 UTC; 1 day 23h ago\n     Docs: man:bluetoothd(8)\n Main PID: 530 (bluetoothd)\n   Status: "Running"\n   CGroup: /system.slice/bluetooth.service\n           \xe2\x94\x94\xe2\x94\x80530 /usr/lib/bluetooth/bluetoothd\n\nOct 13 21:48:58 RPi3 systemd[1]: Starting Bluetooth service...\nOct 13 21:48:58 RPi3 bluetoothd[530]: Bluetooth daemon 5.43\nOct 13 21:48:58 RPi3 systemd[1]: Started Bluetooth service.\nOct 13 21:48:58 RPi3 bluetoothd[530]: Bluetooth management interface 1.14 initialized\nOct 13 21:48:58 RPi3 bluetoothd[530]: Failed to obtain handles for "Service Changed" characteristic\nOct 13 21:48:58 RPi3 bluetoothd[530]: Endpoint registered: sender=:1.10 path=/MediaEndpoint/A2DPSource\nOct 13 21:48:58 RPi3 bluetoothd[530]: Endpoint registered: sender=:1.10 path=/MediaEndpoint/A2DPSink\n'

    def get_bluetooth_service(self, cmd, shell):
        if TestDbusModuleCalls.experimental:
            return TestDbusModuleCalls.bluetooth_service_experimental
        else:
            return TestDbusModuleCalls.bluetooth_service_normal

    def setUp(self):
        """
        Patch the DBus module
        :return:
        """
        self.dbus_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()
        self.process_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
            'subprocess': self.process_mock
        }
        self.dbus_mock.Interface.return_value.GetManagedObjects.return_value = tests.obj_data.full_ubits
        self.process_mock.check_output = self.get_bluetooth_service
        self.process_mock.run.return_value = subprocess.CompletedProcess(
            args=['bluetoothctl', '-v'], returncode=0, stdout=b'bluetoothctl: 5.53\n', stderr=b'')
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import dbus_tools
        self.module_under_test = dbus_tools

    def tearDown(self):
        self.module_patcher.stop()

    def test_uuid_path_gatt(self):
        dbus_full_path = self.module_under_test.get_dbus_path(adapter='00:00:00:00:5A:AD',
                                                              device='F7:17:E4:09:C0:C6',
                                                              service='e95df2d8-251d-470a-a062-fa1922dfa9a8',
                                                              characteristic='e95d9715-251d-470a-a062-fa1922dfa9a8',
                                                              descriptor='00002902-0000-1000-8000-00805f9b34fb')
        expected_result = '/org/bluez/hci0/dev_F7_17_E4_09_C0_C6/service0031/char0035/desc0037'
        self.assertEqual(dbus_full_path, expected_result)

    def test_bad_path(self):
        path_found = self.module_under_test.get_dbus_path(adapter='00:00:00:00:5A:C6',
                                                          device='F7:17:E4:09:C0:XX',
                                                          service='e95df2d8-251d-470a-a062-fa1922dfa9a8')
        expected_path = None
        self.assertEqual(path_found, expected_path)

    def test_get_iface_from_path(self):
        my_iface = self.module_under_test.get_iface(adapter='00:00:00:00:5A:AD',
                                                    device='F7:17:E4:09:C0:C6',
                                                    service='e95df2d8-251d-470a-a062-fa1922dfa9a8',
                                                    characteristic='e95d9715-251d-470a-a062-fa1922dfa9a8',
                                                    descriptor='00002902-0000-1000-8000-00805f9b34fb')
        self.assertEqual(constants.GATT_DESC_IFACE, my_iface)

    def test_profile_path(self):
        my_iface = self.module_under_test.get_profile_path(adapter='00:00:00:00:5A:AD',
                                                           device='F7:17:E4:09:C0:C6',
                                                           profile='e95df2d8-251d-470a-a062-fa1922dfa9a8')
        self.assertEqual(None, my_iface)

    def test_bluez_version(self):
        bluez_ver = self.module_under_test.bluez_version()
        self.assertEqual('5.53', bluez_ver)

    def test_bluez_service_experimental(self):
        TestDbusModuleCalls.experimental = True
        bluez_exper = self.module_under_test.bluez_experimental_mode()
        self.assertTrue(bluez_exper)

    def test_bluez_service_normal(self):
        TestDbusModuleCalls.experimental = False
        bluez_exper = self.module_under_test.bluez_experimental_mode()
        self.assertFalse(bluez_exper)

    def test_get_device_address_from_dbus_path(self):
        """
        Get mac address fromo any given dbus_path that includes
        "dev_xx_xx_xx_xx"
        """
        test_data = [
            ['/org/bluez/hci0/dev_EB_F6_95_27_84_A0', 'EB:F6:95:27:84:A0'],
            ['/org/bluez/hci0', ''],
            ['/org/bluez/hci0/dev_EB_F6_95_27_84_A0/player0',
             'EB:F6:95:27:84:A0']
        ]
        for i in range(0, len(test_data)):
            with self.subTest(i=i):
                self.assertEqual(
                    test_data[i][1],
                    self.module_under_test.get_device_address_from_dbus_path(
                        test_data[i][0]
                    ))

    def test_mac_addr_deprecated(self):
        with patch('logging.Logger.warning') as logger:
            self.module_under_test.get_mac_addr_from_dbus_path(
                '/org/bluez/hci0/dev_EB_F6_95_27_84_A0')
            logger.assert_called_once_with('get_mac_addr_from_dbus_path has '
                                           'been deprecated and has been '
                                           'replaced with '
                                           'get_device_address_from_dbus_path')

    def test_get_device_address(self):
        expected = [{'E4:43:33:7E:54:1C': 'BBC micro:bit [pugit]'}]
        result = self.module_under_test.get_device_addresses('pugit')
        self.assertTrue(isinstance(result, list))
        self.assertDictEqual(expected[0], result[0])


if __name__ == '__main__':
    unittest.main()



import dbus
import dbusmock
import io
from pathlib import Path
import subprocess
# from gi.repository import GLib
from unittest import mock, skip
from tests.mock_eventloop import MockAsync, run_pending_events

from examples import microbit_poll


class TestExampleScanner(dbusmock.DBusTestCase):
    """
    Test mocking bluetoothd
    """

    @classmethod
    def setUpClass(cls):
        here = Path(__file__).parent
        template = str(here.joinpath('dbusmock_templates', 'bluez_scan.py'))
        cls.start_system_bus()
        cls.dbus_con = cls.get_dbus(True)
        (cls.p_mock, cls.obj_bluez) = cls.spawn_server_template(
            template, {}, stdout=subprocess.PIPE)

    def setUp(self):
        self.obj_bluez.Reset()
        self.dbusmock = dbus.Interface(self.obj_bluez, dbusmock.MOCK_IFACE)
        self.dbusmock_bluez = dbus.Interface(self.obj_bluez, 'org.bluez.Mock')

    @classmethod
    def tearDownClass(cls):
        cls.stop_dbus(cls.system_bus_pid)
        cls.p_mock.terminate()
        cls.p_mock.wait()

    def test_add_microbit(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        path = self.dbusmock_bluez.AddDevice('hci0', 'E9:06:4D:45:FC:8D', '')
        dev = self.dbus_con.get_object('org.bluez', path)
        name = dev.Get('org.bluez.Device1', 'Name')
        alias = dev.Get('org.bluez.Device1', 'Alias')
        self.assertIn('micro:bit', name)
        self.assertIn('micro:bit', alias)

    def test_connect_microbit(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        path = self.dbusmock_bluez.AddDevice('hci0', 'E9:06:4D:45:FC:8D', '')
        dev = self.dbus_con.get_object('org.bluez', path)
        self.assertFalse(dev.Get('org.bluez.Device1', 'Connected'))
        self.dbusmock_bluez.ConnectDevice('hci0', 'E9:06:4D:45:FC:8D')
        self.assertTrue(dev.Get('org.bluez.Device1', 'Connected'))

    @skip("Poll test is timing out")
    def test_poll_example(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')
        path = self.dbusmock_bluez.AddDevice('hci0', 'E9:06:4D:45:FC:8D', '')
        # self.dbusmock_bluez.ConnectMicroBit()
        # path = dbus.ObjectPath('/org/bluez/hci0/dev_E9_06_4D_45_FC_8D/service003c')
        # srvc_props = dbus.Dictionary({'UUID': 'e95d6100-251d-470a-a062-fa1922dfa9a8',
        #                               'Device': '/org/bluez/hci0/dev_E9_06_4D_45_FC_8D',
        #                               'Primary': True}, signature='sv')
        # self.dbusmock_bluez.AddGattService(path, srvc_props)
        # path = dbus.ObjectPath('/org/bluez/hci0/dev_E9_06_4D_45_FC_8D/service003c/char0040')
        # chrc_props = dbus.Dictionary({
        #     'UUID': 'e95d1b25-251d-470a-a062-fa1922dfa9a8',
        #     'Service': '/org/bluez/hci0/dev_E9_06_4D_45_FC_8D/service003c',
        #     'Value': [], 'Flags': ['read', 'write']}, signature='sv')
        # chrc_props['Value'] = dbus.Array(chrc_props['Value'], signature='y')
        # self.dbusmock_bluez.AddGattCharacteristic(path, chrc_props)
        with mock.patch('time.sleep', return_value=None):
            microbit_poll.main()



import dbus
import dbusmock
import io
from pathlib import Path
import subprocess
from unittest import mock
from unittest.mock import patch
from tests.mock_eventloop import MockAsync

# Module under test
from bluezero import dbus_tools


class TestAdapterExample(dbusmock.DBusTestCase):
    """
    Test mocking bluetoothd
    """

    @classmethod
    def setUpClass(cls):
        here = Path(__file__).parent
        template = str(here.joinpath('dbusmock_templates', 'bluez_scan.py'))
        cls.start_system_bus()
        cls.dbus_con = cls.get_dbus(True)
        (cls.p_mock, cls.obj_bluez) = cls.spawn_server_template(
            template, {}, stdout=subprocess.PIPE)

    def setUp(self):
        self.obj_bluez.Reset()
        self.dbusmock = dbus.Interface(self.obj_bluez, dbusmock.MOCK_IFACE)
        self.dbusmock_bluez = dbus.Interface(self.obj_bluez, 'org.bluez.Mock')


    @classmethod
    def tearDownClass(cls):
        cls.stop_dbus(cls.system_bus_pid)
        cls.p_mock.terminate()
        cls.p_mock.wait()

    def test_dbus_string(self):
        result = dbus_tools.dbus_to_python(dbus.String("This is a test"))
        self.assertEqual(result, "This is a test")
        self.assertIsInstance(result, str)

    def test_dbus_int64(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')

        result = dbus_tools.dbus_to_python(dbus.Int64(12))
        self.assertTrue(isinstance(result, int))

    def test_dbus_read_value(self):
        result = dbus_tools.dbus_to_python(dbus.Array([dbus.Byte(66),
                                                       dbus.Byte(97),
                                                       dbus.Byte(122)],
                                                      signature=dbus.Signature('y')))
        self.assertEqual(result, b"Baz")

    def test_dbus_dict(self):
        result = dbus_tools.dbus_to_python(dbus.Dictionary(
            {dbus.String('device'): dbus.ObjectPath('/org/bluez/hci0/dev_67_63_13_0D_37_01', variant_level=1),
             dbus.String('link'): dbus.String('LE', variant_level=1),
             dbus.String('mtu'): dbus.UInt16(512, variant_level=1)},
            signature=dbus.Signature('sv'))
        )
        self.assertDictEqual({'device': '/org/bluez/hci0/dev_67_63_13_0D_37_01',
                              'link': 'LE', 'mtu': 512},
                             result)

    def test_dbus_dict_adapter_props(self):
        dongle_props = dbus.Dictionary({
            dbus.String('Address'): dbus.String('11:22:33:44:55:66', variant_level=1),
            dbus.String('AddressType'): dbus.String('public', variant_level=1),
            dbus.String('Name'): dbus.String('test_machine', variant_level=1),
            dbus.String('Alias'): dbus.String('test_machine', variant_level=1),
            dbus.String('Class'): dbus.UInt32(786700, variant_level=1),
            dbus.String('Powered'): dbus.Boolean(True, variant_level=1),
            dbus.String('Discoverable'): dbus.Boolean(False, variant_level=1),
            dbus.String('DiscoverableTimeout'): dbus.UInt32(190, variant_level=1),
            dbus.String('Pairable'): dbus.Boolean(False, variant_level=1),
            dbus.String('PairableTimeout'): dbus.UInt32(120, variant_level=1),
            dbus.String('Discovering'): dbus.Boolean(False, variant_level=1),
            dbus.String('UUIDs'): dbus.Array(
                [dbus.String('0000110e-0000-1000-8000-00805f9b34fb'),
                 dbus.String('0000110a-0000-1000-8000-00805f9b34fb'),
                 dbus.String('00001200-0000-1000-8000-00805f9b34fb'),
                 dbus.String('00001112-0000-1000-8000-00805f9b34fb'),
                 dbus.String('0000110b-0000-1000-8000-00805f9b34fb'),
                 dbus.String('0000110c-0000-1000-8000-00805f9b34fb'),
                 dbus.String('00001800-0000-1000-8000-00805f9b34fb'),
                 dbus.String('00001108-0000-1000-8000-00805f9b34fb'),
                 dbus.String('00001801-0000-1000-8000-00805f9b34fb')],
                signature=dbus.Signature('s'), variant_level=1),
            dbus.String('Modalias'): dbus.String('usb:v1D6Bp0246d0535', variant_level=1)},
            signature=dbus.Signature('sv'))
        expected = {'Address': '11:22:33:44:55:66',
                    'AddressType': 'public',
                    'Alias': 'test_machine',
                    'Class': 786700,
                    'Discoverable': False,
                    'DiscoverableTimeout': 190,
                    'Discovering': False,
                    'Modalias': 'usb:v1D6Bp0246d0535',
                    'Name': 'test_machine',
                    'Pairable': False,
                    'PairableTimeout': 120,
                    'Powered': True,
                    'UUIDs': ['0000110e-0000-1000-8000-00805f9b34fb',
                              '0000110a-0000-1000-8000-00805f9b34fb',
                              '00001200-0000-1000-8000-00805f9b34fb',
                              '00001112-0000-1000-8000-00805f9b34fb',
                              '0000110b-0000-1000-8000-00805f9b34fb',
                              '0000110c-0000-1000-8000-00805f9b34fb',
                              '00001800-0000-1000-8000-00805f9b34fb',
                              '00001108-0000-1000-8000-00805f9b34fb',
                              '00001801-0000-1000-8000-00805f9b34fb']}
        result = dbus_tools.dbus_to_python(dongle_props)
        self.assertEqual(expected, result)

    def test_dbus_bool(self):
        result = dbus_tools.dbus_to_python(dbus.Boolean(True))
        self.assertTrue(isinstance(result, bool))

    def test_dbus_double(self):
        result = dbus_tools.dbus_to_python(dbus.Double(12.3))
        self.assertTrue(12.3, result)

    def test_dbus_byte(self):
        data = dbus.Byte(255)
        result = dbus_tools.dbus_to_python(data)
        self.assertEqual(0xFF, result)

    def test_dbus_byte_array1(self):
        data = dbus.Array([dbus.Byte(29)],
                          signature=dbus.Signature('y'))
        result = dbus_tools.dbus_to_python(data)
        self.assertEqual(b'\x1d', result)

    def test_str_to_dbusarray(self):
        expected = dbus.Array([dbus.Byte(70), dbus.Byte(111), dbus.Byte(120)],
                              signature=dbus.Signature('y'))
        result = dbus_tools.str_to_dbusarray('Fox')
        self.assertEqual(expected, result)

    def test_bytes_to_dbusarray(self):
        expected = dbus.Array([dbus.Byte(70), dbus.Byte(111), dbus.Byte(120)],
                              signature=dbus.Signature('y'))
        result = dbus_tools.bytes_to_dbusarray(b'Fox')
        self.assertEqual(expected, result)

    def test_bytes_to_dbusarray2(self):
        expected = dbus.Array([dbus.Byte(0), dbus.Byte(1), dbus.Byte(2)],
                              signature=dbus.Signature('y'))
        result = dbus_tools.bytes_to_dbusarray(b'\x00\x01\x02')
        self.assertEqual(expected, result)

    def test_get_service(self):
        result = dbus_tools.get_services('some/rubbish')
        self.assertListEqual([], result)

    def test_interfaces_added(self):
        with patch('logging.Logger.debug') as logger:
            result = dbus_tools.interfaces_added('/org/bluez/hci1/dev_00_11_22_33_44_55',
                                                 'org.bluez.Device1')
            logger.assert_called_once()

    def test_properties_changed(self):
        with patch('logging.Logger.debug') as logger:
            result = dbus_tools.properties_changed('org.bluez.Device1',
                                                   {'Connected': True},
                                                   {}, '/org/bluez/hci0')
            print(result)
            logger.assert_called_once()

    def test_get_property(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')

        path_obj = dbus_tools.get_dbus_obj('/org/bluez/hci0')
        result = dbus_tools.get(path_obj, 'org.bluez.Adapter1', 'Address')
        self.assertEqual('00:01:02:03:04:05', result)

    def test_get_property_exception(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')

        path_obj = dbus_tools.get_dbus_obj('/org/bluez/hci0')
        with self.assertRaises(dbus.exceptions.DBusException):
            result = dbus_tools.get(path_obj, 'org.bluez.AdapterXX', 'Address')

    def test_get_property_default(self):
        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')

        path_obj = dbus_tools.get_dbus_obj('/org/bluez/hci0')
        result = dbus_tools.get(path_obj, 'org.bluez.Adapter1', 'address', 'xx')
        self.assertEqual('xx', result)



"""Utility functions for DBus use within Bluezero."""

# Standard libraries
import re
import subprocess

# D-Bus import
import dbus

# python-bluezero constants import
from bluezero import constants
from bluezero import tools

logger = tools.create_module_logger(__name__)


def bluez_version():
    """
    get the version of the BlueZ daemon being used on the system

    :return: String of BlueZ version
    """
    cmd = ['bluetoothctl', '-v']
    cmd_output = subprocess.run(cmd, capture_output=True, check=True)
    version = cmd_output.stdout.decode('utf-8').split()
    return version[1]


def bluez_experimental_mode():
    """
    Return True if the BlueZ daemon service is in experimental mode
    :return: True if experimental enabled
    """
    status = subprocess.check_output('service bluetooth status', shell=True)
    if re.search('--experimental', status.decode('utf-8')) is None:
        return False
    else:
        return True


def interfaces_added(path, interfaces):
    """
    Callback for when an interface is added

    :param path:
    :param interfaces:
    :return:
    """
    if constants.DEVICE_INTERFACE in interfaces:
        logger.debug('Device added at %s', path)


def properties_changed(interface, changed, invalidated, path):
    """
    Callback for when properties are changed

    :param interface:
    :param changed:
    :param invalidated:
    :param path:
    :return:
    """
    if constants.DEVICE_INTERFACE in interface:
        for prop in changed:
            logger.debug('%s:%s Property %s new value %s', interface, path,
                         prop, changed[prop])


def get_dbus_obj(dbus_path):
    """
    Get the the DBus object for the given path
    :param dbus_path:
    :return:
    """
    bus = dbus.SystemBus()
    return bus.get_object(constants.BLUEZ_SERVICE_NAME, dbus_path)


def get_dbus_iface(iface, dbus_obj):
    """
    Return the DBus interface object for given interface and DBus object
    :param iface:
    :param dbus_obj:
    :return:
    """
    return dbus.Interface(dbus_obj, iface)


def get_managed_objects():
    """Return the objects currently managed by the DBus Object Manager."""
    bus = dbus.SystemBus()
    manager = dbus.Interface(bus.get_object(
        constants.BLUEZ_SERVICE_NAME, '/'),
        constants.DBUS_OM_IFACE)
    return manager.GetManagedObjects()


def get_mac_addr_from_dbus_path(path):
    """
    [Deprecated] Function to get the address of a remote device from
    a given D-Bus path. Path must include device part
    (e.g. dev_XX_XX_XX_XX_XX_XX)
    """
    logger.warning('get_mac_addr_from_dbus_path has been deprecated and has '
                   'been replaced with get_device_address_from_dbus_path')
    return get_device_address_from_dbus_path(path)


def get_device_address_from_dbus_path(path):
    """
    [Deprecated] Function to get the address of a remote device from
    a given D-Bus path. Path must include device part
    (e.g. dev_XX_XX_XX_XX_XX_XX)
    """
    for path_elem in path.split('/'):
        if path_elem.startswith('dev_'):
            return path_elem.replace("dev_", '').replace("_", ":")
    return ''


def get_adapter_address_from_dbus_path(path):
    """Return the address of the adapter from the a DBus path"""
    result = re.match(r'/org/bluez/hci\d+', path)
    mngd_objs = get_managed_objects()
    return mngd_objs[result.group(0)][constants.ADAPTER_INTERFACE]['Address']


def _get_dbus_path2(objects, parent_path, iface_in, prop, value):
    """
    Find DBus path for given DBus interface with property of a given value.

    :param objects: Dictionary of objects to search
    :param parent_path: Parent path to include in search
    :param iface_in: The interface of interest
    :param prop: The property to search for
    :param value: The value of the property being searched for
    :return: Path of object searched for
    """
    if parent_path is None:
        return None
    for path, iface in objects.items():
        props = iface.get(iface_in)
        if props is None:
            continue
        dev_name = "dev_" + value.lower().replace(":", "_")
        if (props[prop].lower() == value.lower() or
            path.lower().endswith(dev_name)) \
                and path.startswith(parent_path):
            return path
    return None


def get_dbus_path(adapter=None,
                  device=None,
                  service=None,
                  characteristic=None,
                  descriptor=None):
    """
    Return a DBus path for the given properties
    :param adapter: Adapter address
    :param device: Device address
    :param service: GATT Service UUID
    :param characteristic: GATT Characteristic UUID
    :param descriptor: GATT Descriptor UUID
    :return: DBus path
    """
    bus = dbus.SystemBus()
    manager = dbus.Interface(
        bus.get_object(constants.BLUEZ_SERVICE_NAME, '/'),
        constants.DBUS_OM_IFACE)
    mngd_objs = manager.GetManagedObjects()

    _dbus_obj_path = None

    if adapter is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         '/org/bluez',
                                         constants.ADAPTER_INTERFACE,
                                         'Address',
                                         adapter)

    if device is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         _dbus_obj_path,
                                         constants.DEVICE_INTERFACE,
                                         'Address',
                                         device)

    if service is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         _dbus_obj_path,
                                         constants.GATT_SERVICE_IFACE,
                                         'UUID',
                                         service)

    if characteristic is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         _dbus_obj_path,
                                         constants.GATT_CHRC_IFACE,
                                         'UUID',
                                         characteristic)

    if descriptor is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         _dbus_obj_path,
                                         constants.GATT_DESC_IFACE,
                                         'UUID',
                                         descriptor)
    return _dbus_obj_path


def get_profile_path(adapter,
                     device,
                     profile):
    """
    Return a DBus path for the given properties

    :param adapter: Adapter address
    :param device: Device address
    :param profile:
    :return:
    """
    bus = dbus.SystemBus()
    manager = dbus.Interface(
        bus.get_object(constants.BLUEZ_SERVICE_NAME, '/'),
        constants.DBUS_OM_IFACE)
    mngd_objs = manager.GetManagedObjects()

    _dbus_obj_path = None

    if adapter is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         '/org/bluez',
                                         constants.ADAPTER_INTERFACE,
                                         'Address',
                                         adapter)
    if device is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         _dbus_obj_path,
                                         constants.DEVICE_INTERFACE,
                                         'Address',
                                         device)

    if profile is not None:
        _dbus_obj_path = _get_dbus_path2(mngd_objs,
                                         _dbus_obj_path,
                                         constants.GATT_PROFILE_IFACE,
                                         'UUID',
                                         profile)
    return _dbus_obj_path


def get_iface(adapter=None,
              device=None,
              service=None,
              characteristic=None,
              descriptor=None):
    """
    For the given list of properties return the deepest interface

    :param adapter: Adapter address
    :param device: Device address
    :param service: GATT Service UUID
    :param characteristic: GATT Characteristic UUID
    :param descriptor: GATT Descriptor UUID
    :return: DBus Interface
    """
    if adapter is not None:
        _iface = constants.ADAPTER_INTERFACE

    if device is not None:
        _iface = constants.DEVICE_INTERFACE

    if service is not None:
        _iface = constants.GATT_SERVICE_IFACE

    if characteristic is not None:
        _iface = constants.GATT_CHRC_IFACE

    if descriptor is not None:
        _iface = constants.GATT_DESC_IFACE

    return _iface


def get_methods(adapter=None,
                device=None,
                service=None,
                characteristic=None,
                descriptor=None):
    """
    Get methods available for the specified

    :param adapter: Adapter Address
    :param device: Device Address
    :param service: GATT Service UUID
    :param characteristic: GATT Characteristic UUID
    :param descriptor: GATT Descriptor UUID
    :return: Object of the DBus methods available
    """
    path_obj = get_dbus_path(adapter,
                             device,
                             service,
                             characteristic,
                             descriptor)

    iface = get_iface(adapter,
                      device,
                      service,
                      characteristic,
                      descriptor)
    if path_obj is not None:
        return get_dbus_iface(iface, get_dbus_obj(path_obj))
    else:
        return None


def get_props(adapter=None,
              device=None,
              service=None,
              characteristic=None,
              descriptor=None):
    """
    Get properties for the specified object

    :param adapter: Adapter Address
    :param device:  Device Address
    :param service:  GATT Service UUID
    :param characteristic: GATT Characteristic UUID
    :param descriptor: GATT Descriptor UUID
    :return: Object of the DBus properties available
    """
    path_obj = get_dbus_path(adapter,
                             device,
                             service,
                             characteristic,
                             descriptor)
    if path_obj is not None:
        return get_dbus_iface(dbus.PROPERTIES_IFACE, get_dbus_obj(path_obj))
    else:
        return None


def get_services(path_obj):
    """
    Return a list of GATT Service UUIDs for a given Bluetooth device D-Bus path

    :param path_obj: D-Bus path for remote Bluetooth device
    :return: List of GATT Service UUIDs
    """
    found_services = []
    valid_structure = re.match(r'/org/bluez/hci\d+/dev(_([0-9A-Fa-f]){2}){6}',
                               path_obj)
    if not valid_structure:
        return found_services
    else:
        mngd_obj = get_managed_objects()
        for path in mngd_obj:
            if path.startswith(path_obj):
                srv_uuid = mngd_obj[path].get(
                    constants.GATT_SERVICE_IFACE, {}).get('UUID')
                if srv_uuid:
                    found_services.append(str(srv_uuid))
        return found_services


def get_device_addresses(name_contains):
    """
    Finds device whose name contains the string

    :param name_contains: String to be found in device name
    :return: List of dictionaries wth Device address and name
    """
    found = []
    mngd_obj = get_managed_objects()
    for path in mngd_obj:
        dev_name = mngd_obj[path].get(
            constants.DEVICE_INTERFACE, {}).get('Name')
        if dev_name and name_contains in dev_name:
            dev_addr = mngd_obj[path].get(
                constants.DEVICE_INTERFACE, {}).get('Address')
            found.append({dev_addr: dev_name})
    return found


def get(dbus_prop_obj, dbus_iface, prop_name, default=None):
    """
    Get a property from a D-Bus object and provide a default if the property
    does not exist. Similar to the "get" functionality on a python dictionary.

    :param dbus_prop_obj: The output object from dbus_tools.get_props
    :param dbus_iface: D-Bus Interface name e.g. 'org.bluez.Device1'
    :param prop_name: The name of the property
    :param default: What value to return if the property does not exist.
    :return: The property value if it exists or the default value
    """
    try:
        return dbus_prop_obj.Get(dbus_iface, prop_name)
    except dbus.exceptions.DBusException as dbus_exception:
        err_name = dbus_exception.get_dbus_name()
        err_msg = dbus_exception.get_dbus_message()
        if 'UnknownProperty' in err_name:
            return default
        elif 'no such property' in err_msg.casefold():
            return default
        else:
            raise dbus_exception


def str_to_dbusarray(word):
    """Helper function to represent Python string as D-Dbus Byte array"""
    return dbus.Array([dbus.Byte(ord(letter)) for letter in word], 'y')


def bytes_to_dbusarray(bytesarray):
    """Helper function to represent Python bytearray as D-Bus Byte array"""
    return dbus.Array([dbus.Byte(elem) for elem in bytesarray], 'y')


def dbus_to_python(data):
    """convert dbus data types to python native data types"""
    if isinstance(data, dbus.String):
        data = str(data)
    elif isinstance(data, dbus.Boolean):
        data = bool(data)
    elif isinstance(data, dbus.Byte):
        data = int(data)
    elif isinstance(data, dbus.UInt16):
        data = int(data)
    elif isinstance(data, dbus.UInt32):
        data = int(data)
    elif isinstance(data, dbus.Int64):
        data = int(data)
    elif isinstance(data, dbus.Double):
        data = float(data)
    elif isinstance(data, dbus.ObjectPath):
        data = str(data)
    elif isinstance(data, dbus.Array):
        if data.signature == dbus.Signature('y'):
            data = bytearray(data)
        else:
            data = [dbus_to_python(value) for value in data]
    elif isinstance(data, dbus.Dictionary):
        new_data = dict()
        for key in data:
            new_data[dbus_to_python(key)] = dbus_to_python(data[key])
        data = new_data
    return data



import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants

adapter_props = tests.obj_data.full_ubits


def mock_get(iface, prop):
    return tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop]


def mock_set(iface, prop, value):
    tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop] = value


def mock_register_advertisement(advert_obj, options):
    for prop in advert_obj.props[constants.LE_ADVERTISEMENT_IFACE]:
        tests.obj_data['/org/bluez/hci0']['org.bluez.LEAdvertisingManager1'][prop] = advert_obj[constants.LE_ADVERTISEMENT_IFACE][prop]


class TestBluezeroAdvertisement(unittest.TestCase):

    def setUp(self):
        """
        Patch the DBus module
        :return:
        """
        self.dbus_mock = MagicMock()
        self.dbus_exception_mock = MagicMock()
        self.dbus_service_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.exceptions': self.dbus_exception_mock,
            'dbus.service': self.dbus_service_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
        }
        self.dbus_mock.Interface.return_value.GetManagedObjects.return_value = tests.obj_data.full_ubits
        self.dbus_mock.Interface.return_value.Get = mock_get
        self.dbus_mock.Interface.return_value.Set = mock_set
        self.dbus_mock.return_value
        self.dbus_mock.SystemBus = MagicMock()
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import advertisement
        from bluezero import dbus_tools
        self.module_under_test = advertisement
        self.module_tools = dbus_tools

    def tearDown(self):
        self.module_patcher.stop()

    def test_beacon_default_adapter(self):
        beacon = self.module_under_test.Advertisement(1, 'broadcast')
        beacon.service_UUIDs = ['FEAA']
        beacon.service_data = {'FEAA': [0x10, 0x00, 0x00, 0x63, 0x73,
                                        0x72, 0x00, 0x61, 0x62, 0x6f,
                                        0x75, 0x74]}
        ad_manager = self.module_under_test.AdvertisingManager('00:00:00:00:5A:AD')
        ad_manager.register_advertisement(beacon, {})
        result = self.module_tools.get_managed_objects()['/org/bluez/hci0']['org.bluez.LEAdvertisingManager1']
        self.assertDictEqual({}, result)



"""
Collection of functions to work with the GLib Event Loop
"""
# Main eventloop import
import dbus
import dbus.mainloop.glib
from gi.repository import GLib

from bluezero import tools

logger = tools.create_module_logger(__name__)


def add_timer_ms(time, callback, data=None):
    """Call given callback every x milliseconds"""
    if data:
        GLib.timeout_add(time, callback, data)
    else:
        GLib.timeout_add(time, callback)


def add_timer_seconds(time, callback, data=None):
    """Call given callback every x seconds"""
    if data:
        GLib.timeout_add_seconds(time, callback, data)
    else:
        GLib.timeout_add_seconds(time, callback)


class EventLoop:
    """Facade class to help with using GLib event loop"""
    # def generic_error_cb(self, error):
    #     """Generic Error Callback function."""
    #     logger.error('D-Bus call failed: ' + str(error))
    #     self.mainloop.quit()

    # def __new__(cls):
    #     return object.__new__(cls)

    def __init__(self):
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self.mainloop = GLib.MainLoop()

    def run(self):
        """Run event loop"""
        self.mainloop.run()

    def quit(self):
        """Stop event loop"""
        self.mainloop.quit()

    def is_running(self):
        """Check if event loop is running"""
        self.mainloop.is_running()



"""Classes that represent the GATT features of a remote device."""

from time import sleep

from bluezero import adapter
from bluezero import device
from bluezero import GATT
from bluezero import tools

logger = tools.create_module_logger(__name__)


class Central:
    """Create a BLE instance taking the Central role."""

    def __init__(self, device_addr, adapter_addr=None):
        if adapter_addr is None:
            self.dongle = adapter.Adapter()
            logger.debug('Adapter is: %s', self.dongle.address)
        else:
            self.dongle = adapter.Adapter(adapter_addr)
        if not self.dongle.powered:
            self.dongle.powered = True
            logger.debug('Adapter was off, now powered on')
        self.rmt_device = device.Device(self.dongle.address, device_addr)

        self._characteristics = []

    @staticmethod
    def available(adapter_address=None):
        """Generator for getting a list of devices"""
        return device.Device.available(adapter_address)

    def add_characteristic(self, srv_uuid, chrc_uuid):
        """
        Specify a characteristic of interest on the remote device by using
        the GATT Service UUID and Characteristic UUID

        :param srv_uuid: 128 bit UUID
        :param chrc_uuid: 128 bit UUID
        :return:
        """
        chrc_hndl = GATT.Characteristic(self.dongle.address,
                                        self.rmt_device.address,
                                        srv_uuid,
                                        chrc_uuid)
        self._characteristics.append(chrc_hndl)
        return chrc_hndl

    def load_gatt(self):
        """
        Once the remote device has been connected to and the GATT database
        has been resolved then it needs to be loaded.

        :return:
        """
        for chrc in self._characteristics:
            available = chrc.resolve_gatt()
            if available:
                logger.info('Service: %s and characteristic: %s added',
                            chrc.srv_uuid, chrc.chrc_uuid)
            else:
                logger.warning('Service: %s and characteristic: %s not '
                               'available on device: %s', chrc.srv_uuid,
                               chrc.chrc_uuid, chrc.device_addr)

    @property
    def services_available(self):
        """Get a list of Service UUIDs available on this device"""
        return self.rmt_device.services_available

    @property
    def services_resolved(self):
        """Indicate whether or not service discovery has been resolved"""
        return self.rmt_device.services_resolved

    @property
    def connected(self):
        """Indicate whether the remote device is currently connected."""
        return self.rmt_device.connected

    def connect(self, profile=None, timeout=35):
        """
        Initiate a connection to the remote device and load
        GATT database once resolved

        :param profile: (optional) profile to use for the connection.
        :param timeout: (optional) seconds to wait for connection.
        """
        if profile is None:
            self.rmt_device.connect(timeout=timeout)
        else:
            self.rmt_device.connect(profile, timeout=timeout)
        while not self.rmt_device.services_resolved:
            sleep(0.5)
        self.load_gatt()

    def disconnect(self):
        """Disconnect from the remote device."""
        self.rmt_device.disconnect()

    def run(self):
        """Start event loop"""
        self.dongle.run()

    def quit(self):
        """Stop event loop"""
        self.dongle.quit()



import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants


class TestDbusModuleCalls(unittest.TestCase):
    """
    Testing things that use the Dbus module
    """
    def setUp(self):
        """
        Patch the DBus module
        :return:
        """
        self.dbus_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()
        self.process_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
            'subprocess': self.process_mock
        }
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import async_tools
        self.module_under_test = async_tools.EventLoop()

    def tearDown(self):
        self.module_patcher.stop()

    def test_call_run(self):
        self.module_under_test.run()

    def test_call_quit(self):
        self.module_under_test.quit()



import dbus
import dbusmock
import io
from pathlib import Path
import subprocess
from unittest import mock
from tests.mock_eventloop import MockAsync

# Module under test
from examples import adapter_example


class TestAdapterExample(dbusmock.DBusTestCase):
    """
    Test mocking bluetoothd
    """

    @classmethod
    def setUpClass(cls):
        here = Path(__file__).parent
        template = str(here.joinpath('dbusmock_templates', 'bluez_scan.py'))
        cls.start_system_bus()
        cls.dbus_con = cls.get_dbus(True)
        (cls.p_mock, cls.obj_bluez) = cls.spawn_server_template(
            template, {}, stdout=subprocess.PIPE)

    def setUp(self):
        self.obj_bluez.Reset()
        self.dbusmock = dbus.Interface(self.obj_bluez, dbusmock.MOCK_IFACE)
        self.dbusmock_bluez = dbus.Interface(self.obj_bluez, 'org.bluez.Mock')


    @classmethod
    def tearDownClass(cls):
        cls.stop_dbus(cls.system_bus_pid)
        cls.p_mock.terminate()
        cls.p_mock.wait()

    def test_on_device_found(self):

        class ForTest:
            found_address = None

            @classmethod
            def new_dev(cls, device):
                cls.found_address = device.address

        self.dbusmock_bluez.AddAdapter('hci0', 'My-Test-Device')

        with mock.patch('bluezero.async_tools.EventLoop', MockAsync):
            with mock.patch('sys.stdout', new=io.StringIO()) as fake_out:
                adapter_example.main()
                self.assertIn('address:  00:01:02:03:04:05',
                              fake_out.getvalue())



"""Classes that represent the GATT features of a remote device."""

import dbus

from bluezero import constants
from bluezero import dbus_tools
from bluezero import device
from bluezero import tools


logger = tools.create_module_logger(__name__)


class Service:
    """Remote GATT Service."""

    def __init__(self, adapter_addr, device_addr, srv_uuid):
        """
        Remote GATT Service Initialisation.

        :param adapter_addr: Adapter address.
        :param device_addr: device address.
        :param srv_uuid: Service UUID.
        """
        self.adapter_addr = adapter_addr
        self.device_addr = device_addr
        self.srv_uuid = srv_uuid
        self.rmt_device = device.Device(adapter_addr, device_addr)
        self.service_methods = None
        self.service_props = None

        if self.rmt_device.services_resolved:
            self.resolve_gatt()

    def resolve_gatt(self):
        """
        Get the methods and properties for the discovered Services

        :return:
        """
        if self.rmt_device.services_resolved:
            self.service_methods = dbus_tools.get_methods(self.adapter_addr,
                                                          self.device_addr,
                                                          self.srv_uuid)
            self.service_props = dbus_tools.get_props(self.adapter_addr,
                                                      self.device_addr,
                                                      self.srv_uuid)

    @property
    def UUID(self):  # pylint: disable=invalid-name
        """
        Return the value of the Service UUID for this path.

        :return: string for example '00001800-0000-1000-8000-00805f9b34fb'
        """
        return self.service_props.Get(
            constants.GATT_SERVICE_IFACE, 'UUID')

    @property
    def device(self):
        """
        Return the DBus object that this service belongs to.

        :return: DBus object of device
        """
        return self.service_props.Get(
            constants.GATT_SERVICE_IFACE, 'Device')

    @property
    def primary(self):
        """
        Return a boolean saying if this a primary service.

        :return: boolean
        """
        return self.service_props.Get(
            constants.GATT_SERVICE_IFACE, 'Primary')


class Characteristic:
    """Remote GATT Characteristic."""

    def __init__(self, adapter_addr, device_addr, srv_uuid, chrc_uuid):
        """
        Remote GATT Characteristic Initialisation.

        :param adapter_addr: Adapter address.
        :param device_addr: device address.
        :param srv_uuid: Service UUID.
        :param chrc_uuid: Characteristic UUID.
        """
        self.adapter_addr = adapter_addr
        self.device_addr = device_addr
        self.rmt_device = device.Device(adapter_addr, device_addr)
        self.srv_uuid = srv_uuid
        self.chrc_uuid = chrc_uuid
        self.characteristic_methods = None
        self.characteristic_props = None
        self._prop_chngd_sig = None

    def resolve_gatt(self):
        """
        Get the methods and properties for the discovered characteristics

        :return: Boolean of if characteristics have been resolved
        """
        logger.info('Resolving GATT database for %s', self.chrc_uuid)
        if self.rmt_device.services_resolved:
            self.characteristic_methods = dbus_tools.get_methods(
                self.adapter_addr,
                self.device_addr,
                self.srv_uuid,
                self.chrc_uuid)
            self.characteristic_props = dbus_tools.get_props(
                self.adapter_addr,
                self.device_addr,
                self.srv_uuid,
                self.chrc_uuid)
            return True
        return False

    @property
    def UUID(self):  # pylint: disable=invalid-name
        """
        Return the value of the Characteristic UUID for this path.

        :return: string example '00002a00-0000-1000-8000-00805f9b34fb'
        """
        return self.characteristic_props.Get(
            constants.GATT_CHRC_IFACE, 'UUID')

    @property
    def service(self):
        """
        Return the DBus object for this characteristic's service.

        :return: DBus object of device
        """
        return self.characteristic_props.Get(
            constants.GATT_CHRC_IFACE, 'Service')

    @property
    def value(self):
        """
        The cached value of the characteristic.

        This property gets updated only after a successful read request and
        when a notification or indication is received, upon which a
        PropertiesChanged signal will be emitted.

        :return: DBus byte array
        """
        return self.read_raw_value()

    @value.setter
    def value(self, new_value):
        if not isinstance(new_value, list):
            new_value = [new_value]
        self.write_value(new_value)

    @property
    def notifying(self):
        """
        Return whether this characteristic has notifications enabled.

        :return: Boolean
        """
        return self.characteristic_props.Get(
            constants.GATT_CHRC_IFACE, 'Notifying')

    @property
    def flags(self):
        """
        Return a list of how this characteristic's value can be used.

        :return: list example ['read', 'write', 'notify']
        """
        return self.characteristic_props.Get(
            constants.GATT_CHRC_IFACE, 'Flags')

    def read_raw_value(self, flags=''):
        """
        Return this characteristic's value (if allowed).

        :param flags: "offset": Start offset
                      "device": Device path (Server only)
        :return:

        Possible Errors:    org.bluez.Error.Failed
                            org.bluez.Error.InProgress
                            org.bluez.Error.NotPermitted
                            org.bluez.Error.InvalidValueLength
                            org.bluez.Error.NotAuthorized
                            org.bluez.Error.NotSupported
        """
        try:
            return self.characteristic_methods.ReadValue(dbus.Array())
        except AttributeError:
            logger.error('Service: %s with Characteristic: %s not defined on '
                         'device: %s', self.srv_uuid, self.chrc_uuid,
                         self.device_addr)
            return []

    def write_value(self, value, flags=''):
        """
        Write a new value to the characteristic.

        :param value: A list of byte values
        :param flags: Optional dictionary.
            Typically empty. Values defined at:
            https://git.kernel.org/pub/scm/bluetooth/bluez.git/tree/doc/gatt-api.txt
        """
        try:
            self.characteristic_methods.WriteValue(value, dbus.Array(flags))
        except AttributeError:
            logger.error('Service: %s with Characteristic: %s not defined '
                         'on device: %s. Cannot write_value',  self.srv_uuid,
                         self.chrc_uuid, self.device_addr)

    def start_notify(self):
        """Initialise notifications for this characteristic."""
        try:
            self.characteristic_methods.StartNotify(
                reply_handler=self.start_notify_cb,
                error_handler=generic_error_cb,
                dbus_interface=constants.GATT_CHRC_IFACE)
        except AttributeError:
            logger.error('Service: %s with Characteristic: %s not defined '
                         'on device: %s. Cannot start_notify',  self.srv_uuid,
                         self.chrc_uuid, self.device_addr)

    def stop_notify(self):
        """Stop notifications for this characteristic."""
        try:
            self.characteristic_methods.StopNotify(
                reply_handler=self.stop_notify_cb,
                error_handler=generic_error_cb,
                dbus_interface=constants.GATT_CHRC_IFACE)
        except AttributeError:
            logger.error('Service: %s with Characteristic: %s not defined on'
                         'on device: %s', self.srv_uuid, self.chrc_uuid,
                         self.device_addr)

    def add_characteristic_cb(self, callback=None):
        """
        Add a callback for this characteristic.

        :param callback: callback function to be added.
        """
        if callback is None:
            self._prop_chngd_sig.remove()
        else:
            self._prop_chngd_sig = self.characteristic_props.connect_to_signal(
                'PropertiesChanged', callback)

    def props_changed_cb(self, iface, changed_props, invalidated_props):
        """
        Callback indicating that properties have changed.

        :param iface: Interface associated with the callback.
        :param changed_props: Properties changed that triggered the callback.
        :param invalidated_props: Unused.
        """
        if iface != constants.GATT_CHRC_IFACE:
            return

        if not changed_props:
            return

        value = changed_props.get('Value', None)
        if not value:
            return

        logger.debug('Properties changed: ', value)

    def start_notify_cb(self):
        """Callback associated with enabling notifications."""
        logger.info('Notifications enabled')

    def stop_notify_cb(self):
        """Callback associated with disabling notifications."""
        logger.info('Notifications disabled')


class Descriptor:
    """Remote GATT Descriptor."""

    def __init__(self, adapter_addr, device_addr,
                 srv_uuid, chrc_uuid, dscr_uuid):
        """
        Remote GATT Descriptor Initialisation.

        :param adapter_addr: Adapter address.
        :param device_addr: device address.
        :param srv_uuid: Service UUID.
        :param chrc_uuid: Characteristic UUID.
        :param dscr_uuid: Descriptor UUID.
        """
        self.adapter_addr = adapter_addr
        self.device_addr = device_addr
        self.rmt_device = device.Device(adapter_addr, device_addr)
        self.srv_uuid = srv_uuid
        self.chrc_uuid = chrc_uuid
        self.dscr_uuid = dscr_uuid
        self.descriptor_methods = None
        self.descriptor_props = None

        if self.rmt_device.services_resolved:
            self.resolve_gatt()

    @property
    def UUID(self):  # pylint: disable=invalid-name
        """
        Return the value of the Descriptor UUID for this path.

        :return: string example '00002a00-0000-1000-8000-00805f9b34fb'
        """
        return self.descriptor_props.Get(constants.GATT_DESC_IFACE, 'UUID')

    @property
    def characteristic(self):
        """
        Object path of the GATT characteristic's descriptor.

        :return: DBus object
        """
        return self.descriptor_props.Get(constants.GATT_DESC_IFACE, 'UUID')

    @property
    def value(self):
        """
        The cached value of the descriptor.

        This property gets updated only after a successful read request, upon
        which a PropertiesChanged signal will be emitted.

        :return: DBus byte array
        """
        return self.descriptor_props.Get(
            constants.GATT_CHRC_IFACE, 'Value')

    @property
    def flags(self):
        """
        Return a list of how this descriptor value can be used.

        :return: list example ['read', 'write']
        """
        return self.descriptor_props.Get(
            constants.GATT_CHRC_IFACE, 'Flags')

    def read_raw_value(self, flags=''):
        """
        Issue a request to read the value of the descriptor.

        Returns the value if the operation was successful.

        :param flags: "offset": Start offset
                      "device": Device path (Server only)

        :return: dbus byte array
        """
        return self.descriptor_methods.ReadValue(dbus.Array(flags))

    def write_value(self, value, flags=''):
        """
        Issue a request to write the value of the descriptor.

        :param value: DBus byte array
        :param flags: "offset": Start offset
                      "device": Device path (Server only)

        :return:
        """
        self.descriptor_methods.WriteValue(value, dbus.Array(flags))


class Profile:
    """Remote GATT Profile."""

    def __init__(self, adapter_addr, device_addr, profile_uuid):
        """
        Remote GATT Profile Initialisation.

        :param profile_path: dbus path to the profile.
        """
        self.profile_path = dbus_tools.get_profile_path(adapter_addr,
                                                        device_addr,
                                                        profile_uuid)
        self.bus = dbus.SystemBus()
        self.profile_object = self.bus.get_object(
            constants.BLUEZ_SERVICE_NAME,
            self.profile_path)
        self.profile_methods = dbus.Interface(
            self.profile_object,
            constants.GATT_PROFILE_IFACE)
        self.profile_props = dbus.Interface(self.profile_object,
                                            dbus.PROPERTIES_IFACE)

    def release(self):
        """
        Release the profile.

        :return:
        """
        self.profile_methods.Release()

    @property
    def UUIDs(self):  # pylint: disable=invalid-name
        """
        128-bit GATT service UUIDs to auto connect.

        :return: list of UUIDs
        """
        return self.profile_props.Get(constants.GATT_PROFILE_IFACE, 'UUIDs')


def generic_error_cb(error):
    """Generic Error Callback function."""
    logger.error('D-Bus call failed: %s', str(error))


def register_app_cb():
    """Application registration callback."""
    logger.info('GATT application registered')


def register_app_error_cb(error):
    """Application registration error callback."""
    logger.warning('Failed to register application: %s', str(error))
    # mainloop.quit()


class GattManager:
    """GATT Manager."""

    def __init__(self, adapter_addr):
        """
        GATT Manager Initialisation.

        :param manager_path: dbus path to the GATT Manager.
        """
        self.manager_path = dbus_tools.get_dbus_path(adapter_addr)
        self.bus = dbus.SystemBus()
        self.manager_obj = self.bus.get_object(
            constants.BLUEZ_SERVICE_NAME,
            self.manager_path)
        self.manager_methods = dbus.Interface(
            self.manager_obj,
            constants.GATT_MANAGER_IFACE)
        self.manager_props = dbus.Interface(self.manager_obj,
                                            dbus.PROPERTIES_IFACE)

    def register_application(self, application, options):
        """
        Register an application with the GATT Manager.

        :param application: Application object.
        :param options:
        :return:
        """
        self.manager_methods.RegisterApplication(
            application.get_path(),
            dbus.Dictionary(options, signature='sv'),
            reply_handler=register_app_cb,
            error_handler=register_app_error_cb
        )

    def unregister_application(self, application):
        """
        Unregister an application with the GATT Manager.

        :param application: Application object.
        :return:
        """
        self.manager_methods.UnregisterApplication(application)



"""
Class and methods that represent LE Advertising.
This class requires BlueZ to have the experimental flag enabled

Classes:

- Advertisement -- Specifies the Advertisement Data to be broadcast
- AdvertisingManager -- register Advertisement Data which should be
  broadcast to devices
"""
import dbus
import dbus.exceptions
import dbus.mainloop.glib
import dbus.service

from bluezero import constants
from bluezero import dbus_tools
from bluezero import async_tools
from bluezero import adapter
from bluezero import tools

logger = tools.create_module_logger(__name__)

dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)


########################################
# Exception classes
#######################################
class InvalidArgsException(dbus.exceptions.DBusException):
    """This is a D-Bus exception class for Invalid Arguments.

    All this class does is set the internal variable ``_dbus_error_name`` to
    the object path for D-Bus Invalid Argument Exceptions.

    """

    _dbus_error_name = 'org.freedesktop.DBus.Error.InvalidArgs'


class Advertisement(dbus.service.Object):
    """Advertisement data to broadcast Class.

    An example of an Eddystone beacon:
    :Example:

    >>> from bluezero import advertisement
    >>> beacon = advertisement.Advertisement(1, 'broadcast')
    >>> beacon.service_UUIDs = ['FEAA']
    >>> beacon.service_data =  {'FEAA': [0x10, 0x08, 0x03, 0x75, 0x6B,
    >>>                                  0x42, 0x61, 0x7A, 0x2e, 0x67,
    >>>                                  0x69, 0x74, 0x68, 0x75, 0x62,
    >>>                                  0x2E, 0x69, 0x6F]}
    >>> ad_manager = advertisement.AdvertisingManager()
    >>> ad_manager.register_advertisement(beacon, {})
    >>> beacon.start()

    """

    def __init__(self, advert_id, ad_type):
        """Default initialiser.

        Creates the interface to the specified advertising data.
        The DBus path must be specified.

        :param advert_id: Unique ID of advertisement.
        :param ad_type: Possible values: "broadcast" or "peripheral"
        """
        # Setup D-Bus object paths and register service
        self.path = '/ukBaz/bluezero/advertisement{0:04d}'.format(advert_id)
        self.bus = dbus.SystemBus()
        self.mainloop = async_tools.EventLoop()
        self.interface = constants.LE_ADVERTISEMENT_IFACE
        dbus.service.Object.__init__(self, self.bus, self.path)
        self.props = {
            constants.LE_ADVERTISEMENT_IFACE: {
                'Type': ad_type,
                'ServiceUUIDs': None,
                'ManufacturerData': None,
                'SolicitUUIDs': None,
                'ServiceData': None,
                'Includes': set(),
                'Appearance': None,
                'LocalName': None
            }
        }

    def start(self):
        """Start GLib event loop"""
        self.mainloop.run()

    def stop(self):
        """Stop GLib event loop"""
        self.mainloop.quit()

    def get_path(self):
        """Return the DBus object path"""
        return dbus.ObjectPath(self.path)

    @dbus.service.method(constants.LE_ADVERTISEMENT_IFACE,
                         in_signature='',
                         out_signature='')
    def Release(self):  # pylint: disable=invalid-name
        """
        This method gets called when the service daemon
        removes the Advertisement. A client can use it to do
        cleanup tasks. There is no need to call
        UnregisterAdvertisement because when this method gets
        called it has already been unregistered.

        :return:
        """
        pass

    @property
    def service_UUIDs(self):  # pylint: disable=invalid-name
        """List of UUIDs that represent available services."""
        return self.Get(constants.LE_ADVERTISEMENT_IFACE,
                        'ServiceUUIDs')

    @service_UUIDs.setter
    def service_UUIDs(self, UUID):  # pylint: disable=invalid-name
        self.Set(constants.LE_ADVERTISEMENT_IFACE,
                 'ServiceUUIDs',
                 UUID)

    def manufacturer_data(self, company_id, data):
        """Manufacturer Data to be broadcast"""
        return self.Set(constants.LE_ADVERTISEMENT_IFACE,
                        'ManufacturerData',
                        {company_id: dbus.Array(data, signature='y')})

    def solicit_UUIDs(self):  # pylint: disable=invalid-name
        """Manufacturer Data to be broadcast (Currently not supported)"""
        pass

    @property
    def service_data(self):
        """Service Data to be broadcast"""
        return self.Get(constants.LE_ADVERTISEMENT_IFACE,
                        'ServiceData')

    @service_data.setter
    def service_data(self, data):
        for uuid in data:
            self.Set(constants.LE_ADVERTISEMENT_IFACE,
                     'ServiceData',
                     {uuid: dbus.Array(data[uuid], signature='y')})

    @property
    def include_tx_power(self):
        """Include TX power in advert (Different from beacon packet)"""
        includes = self.Get(constants.LE_ADVERTISEMENT_IFACE, 'Includes')
        return 'tx-power' in includes

    @include_tx_power.setter
    def include_tx_power(self, state):
        if state:
            self.props[
                constants.LE_ADVERTISEMENT_IFACE][
                'Includes'].add('tx-power')
        else:
            self.props[
                constants.LE_ADVERTISEMENT_IFACE][
                'Includes'].discard('tx-power')

    @property
    def local_name(self):
        """Local name of the device included in Advertisement."""
        return self.Get(constants.LE_ADVERTISEMENT_IFACE, 'LocalName')

    @local_name.setter
    def local_name(self, name):
        self.Set(constants.LE_ADVERTISEMENT_IFACE, 'LocalName', name)

    @property
    def appearance(self):
        """Appearance to be used in the advertising report."""
        return self.Get(constants.LE_ADVERTISEMENT_IFACE, 'Appearance')

    @appearance.setter
    def appearance(self, appearance):
        self.Set(constants.LE_ADVERTISEMENT_IFACE, 'Appearance', appearance)

    @dbus.service.method(constants.DBUS_PROP_IFACE,
                         in_signature='s',
                         out_signature='a{sv}')
    def GetAll(self, interface_name):  # pylint: disable=invalid-name
        """Return the advertisement properties.

        This method is registered with the D-Bus at
        ``org.freedesktop.DBus.Properties``

        :param interface: interface to get the properties of.

        The interface must be ``org.bluez.LEAdvertisement1`` otherwise an
        exception is raised.
        """
        if interface_name != constants.LE_ADVERTISEMENT_IFACE:
            raise InvalidArgsException()

        response = dict()
        response['Type'] = self.props[interface_name]['Type']
        if self.props[interface_name]['ServiceUUIDs'] is not None:
            response['ServiceUUIDs'] = dbus.Array(
                self.props[interface_name]['ServiceUUIDs'],
                signature='s')
        if self.props[interface_name]['ServiceData'] is not None:
            response['ServiceData'] = dbus.Dictionary(
                self.props[interface_name]['ServiceData'],
                signature='sv')
        if self.props[interface_name]['ManufacturerData'] is not None:
            response['ManufacturerData'] = dbus.Dictionary(
                self.props[interface_name]['ManufacturerData'],
                signature='qv')
        if self.props[interface_name]['SolicitUUIDs'] is not None:
            response['SolicitUUIDs'] = dbus.Array(
                self.props[interface_name]['SolicitUUIDs'],
                signature='s')
        if self.props[interface_name]['LocalName'] is not None:
            response['LocalName'] = dbus.String(
                    self.props[interface_name]['LocalName'])
        if self.props[interface_name]['Appearance'] is not None:
            response['Appearance'] = dbus.UInt16(
                    self.props[interface_name]['Appearance'])
        response['Includes'] = dbus.Array(
            self.props[interface_name]['Includes'], signature='s')

        return response

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ss', out_signature='v')
    def Get(self, interface_name,  # pylint: disable=invalid-name
            property_name):
        """DBus API for getting a property value"""

        if interface_name != constants.LE_ADVERTISEMENT_IFACE:
            raise InvalidArgsException()

        try:
            return self.GetAll(interface_name)[property_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=interface_name + '.UnknownProperty')

    @dbus.service.method(dbus.PROPERTIES_IFACE,
                         in_signature='ssv', out_signature='')
    def Set(self,  # pylint: disable=invalid-name
            interface_name, property_name, value):
        """Standard D-Bus API for setting a property value"""

        try:
            iface_props = self.props[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                'no such interface ' + interface_name,
                name=self.interface + '.UnknownInterface')

        if property_name not in iface_props:
            raise dbus.exceptions.DBusException(
                'no such property ' + property_name,
                name=self.interface + '.UnknownProperty')

        iface_props[property_name] = value


def register_ad_cb():
    """Advertisement registration callback."""
    print('Advertisement registered')


def register_ad_error_cb(error):
    """Advertisement registration error callback."""
    print('Failed to register advertisement: ' + str(error))


class AdvertisingManager:
    """
    Associate the advertisement to an adapter.
    If no adapter specified then first adapter in list is used
    """

    def __init__(self, adapter_addr=None):

        self.bus = dbus.SystemBus()

        if adapter_addr is None:
            adapters = list(adapter.Adapter.available())
            if len(adapters) > 0:
                use_adapter = adapters[0]
                adapter_addr = use_adapter.address
        else:
            use_adapter = adapter.Adapter(adapter_addr)

        if not use_adapter.discoverable:
            use_adapter.discoverable = True
        self.advert_mngr_path = dbus_tools.get_dbus_path(adapter=adapter_addr)
        self.advert_mngr_obj = self.bus.get_object(
            constants.BLUEZ_SERVICE_NAME,
            self.advert_mngr_path)
        self.advert_mngr_methods = dbus.Interface(
            self.advert_mngr_obj,
            constants.LE_ADVERTISING_MANAGER_IFACE)
        self.advert_mngr_props = dbus.Interface(self.advert_mngr_obj,
                                                dbus.PROPERTIES_IFACE)

    def register_advertisement(self, advertisement, options=dbus.Array()):
        """
        Registers an advertisement object to be sent over the LE
        Advertising channel

        :param advertisement: Advertisement object
        :param options:
        :return:
        """
        self.advert_mngr_methods.RegisterAdvertisement(
            advertisement.path,
            dbus.Dictionary(options, signature='sv'),
            reply_handler=register_ad_cb,
            error_handler=register_ad_error_cb
        )

    def unregister_advertisement(self, advertisement):
        """This unregisters the services that has been
            previously registered. The advertisement object
            must match the  value that was used
            on registration

        :param advertisement:
        :return:
        """
        self.advert_mngr_methods.UnregisterAdvertisement(
            advertisement.path
        )



import unittest
from unittest.mock import MagicMock
from unittest.mock import patch
import tests.obj_data
from bluezero import constants

adapter_props = tests.obj_data.full_ubits


def mock_get(iface, prop):
    return tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop]


def mock_set(iface, prop, value):
    tests.obj_data.full_ubits['/org/bluez/hci0'][iface][prop] = value


class TestBluezeroBroadcaster(unittest.TestCase):
    def setUp(self):
        """
        Patch the DBus module
        :return:
        """
        self.dbus_mock = MagicMock()
        self.dbus_exception_mock = MagicMock()
        self.dbus_service_mock = MagicMock()
        self.mainloop_mock = MagicMock()
        self.gobject_mock = MagicMock()

        modules = {
            'dbus': self.dbus_mock,
            'dbus.exceptions': self.dbus_exception_mock,
            'dbus.service': self.dbus_service_mock,
            'dbus.mainloop.glib': self.mainloop_mock,
            'gi.repository': self.gobject_mock,
        }
        self.dbus_mock.Interface.return_value.GetManagedObjects.return_value = tests.obj_data.full_ubits
        self.dbus_mock.Interface.return_value.Get = mock_get
        self.dbus_mock.Interface.return_value.Set = mock_set
        self.dbus_mock.SystemBus = MagicMock()
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from bluezero import broadcaster
        from bluezero import adapter
        self.bz_adapter = adapter
        self.module_under_test = broadcaster
        self.path = '/org/bluez/hci0/dev_D4_AE_95_4C_3E_A4'
        self.adapter_path = '/org/bluez/hci0'
        self.dev_name = 'BBC micro:bit [zezet]'
        self.address = 'D4:AE:95:4C:3E:A4'

    def tearDown(self):
        self.module_patcher.stop()

    def test_beacon_default_adapter(self):
        my_beacon = self.module_under_test.Beacon()
    #
    # def test_beacon_specified_adapter(self):
    #     beacon2 = self.module_under_test.Beacon('00:00:00:00:5A:AD')

    # def test_service_data(self):
    #     my_beacon = self.module_under_test.Beacon()
    #     my_beacon.add_service_data('FEAA': [0x10, 0x00, 0x00, 0x63, 0x73,
    #                                            0x72, 0x00, 0x61, 0x62, 0x6f,
    #                                            0x75, 0x74])
    #     self.assertEqual(my_beacon.serivce)
